#------------------------------------------------------------------
# $Copyright: Copyright (c) 2018 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/SAPNW/SAPNW.pm,v 1.24 2014/10/22 07:27:58 lpatil Exp $ 
#
# $Author: lpatil $
# $Date: 2014/10/22 07:27:58 $
# $Revision: 1.24 $
#
# Provides:                            SAPNW
# Author:                              vchappav
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPNW;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses
use Hares;          # This contains the definition Hares related  function .

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( uname );
use POSIX qw ( strftime );
use File::Path qw ( mkpath );
use File::stat;
use File::Basename;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
   $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPNW.pm,v 1.24 2014/10/22 07:27:58 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );
my $dirname = "";

BEGIN {
  $dirname = dirname($0);
  if ($dirname =~ /^(.+)$/) {
    $dirname = $1;     # $dirname now untainted
  }
  unshift (@INC, $dirname);
}

my $SYSTEMCTL = "/usr/bin/systemctl";
my $GREP = "/bin/grep";
#-----------------
# Define globals..
#-----------------
my ( $iSapStartSrv, $iCCMSX, $iCCMSR, $bSapStart,$bSapStartSrv, @liRunningProcs ) = (); # Global..

#--------------------------------
# Define SAPNW subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String SAPAdmin,
                              IN String InstProfile,
                              IN String InstType,
                              IN String ProcMon,
                              IN String EnqSrvResName,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram,
                              IN Bool StopService );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry point name
 $localEnvFile                  Absolute path to environment file to source
 $localSAPAdmin                 SAP System Administrator
 $localInstProfile              Instance profile file for the SAP Instance
 $localInstType                 SAP Instance Type
 $localProcMon                  List of processes to be monitored
 $localEnqSrvResName            VCS resource name for Enqueue Server
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 $localStopService              Specifies whether SAP Start Service should be stopped or not.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localSAPAdmin,
       $localInstProfile,
       $localInstType,
       $localProcMon,
       $localEnqSrvResName,
       $localSecondLevelMonitor,
       $localMonitorProgram,
       $localStopService,
       $localUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  SAPAdmin                 [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  EnqSrvResName            [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]
  StopService              [%s]
  UseSystemD               [%s]",

  $sMyName,   $sEntryPointName,   $localEnvFile,    $localSAPAdmin,
  $localInstProfile,         $localInstType,         $localProcMon,
  $localEnqSrvResName,                    $localSecondLevelMonitor,
  $localMonitorProgram ,$localStopService,  $localUseSystemD );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  EnvFile  SAPAdmin  InstProfile  InstType  ProcMon
       EnqSrvResName SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile SAPAdmin InstProfile InstType ProcMon );
  my @lsOfflineArgs = qw ( SAPAdmin InstProfile InstType );
  my @lsMonitorArgs = qw ( SAPAdmin InstProfile InstType ProcMon );
  my @lsCleanArgs   = qw ( SAPAdmin InstProfile InstType );
  my @lsImfRegisterArgs = qw ( SAPAdmin InstProfile InstType ProcMon );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
                    'imf_register'   => \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..


  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  my ($sSAPSID, $sInstName, $sSAPHost) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $localInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

  #-------------------------------------------------------
  # Validation for  offline monitor entry point(s)..
  #-------------------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline|monitor)$/o ) {
    #------------------------------
    # Validate parameter SAPAdmin..
    #------------------------------
    if ( ! defined getpwnam $localSAPAdmin ) {
      &logError ( 20003, "<%s> SAPAdmin [%s] is not a valid UNIX login",
      $sMyName, $localSAPAdmin );
      $iReturnCode = 0; # Failure..
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20004, "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }
    elsif (( $localUseSystemD)              &&        # UseSystemD is set
           ( SystemD::DetectSystemd())      &&        # SystemD is detected
           ($localEnvFile =~ /\S/ )      &&        # EnvFile is specified 
           ($::gsSystemDEnvFile !~ /\S/)) {           # SystemD-compliant EnvFile is not specified
      &logWarn (20456, "<%s> EnvFile[%s] is specified, but SystemD-compliant environment file is not specified in attribute SystemDEnvFile. 
                      The agent will use [%s] as the value for directive EnvironmentFile in the service file.",
                      $sMyName, $localEnvFile, $localEnvFile);
    }

    #---------------------------------------
    # Validate SAPHost..
    #---------------------------------------
    if ( ! inet_aton ( $sSAPHost ) ) {
      &logError ( 20005,
      "<%s> InstProfile [%s] not terminated with valid hostname [%s]",
      $sMyName, $localInstProfile, $sSAPHost );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------
    # Validate parameter InstProfile..
    #---------------------------------
    if ( ! -e $localInstProfile ) {
      &logError ( 20006, "<%s> InstProfile [%s] file does not exist",
      $sMyName, $localInstProfile );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20007, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

    #----------------------------------
    # Check if EnqSrvResName is valid..
    #----------------------------------
    if ( $localEnqSrvResName =~ m/\S+/ && $localInstType =~ /^enqrep$/io  &&
         ! SAPNW::ChkResName ( $localEnqSrvResName,
                                    $sSAPSID,
                                    $localInstType ) ) {
      &logError ( 20008,
      "<%s> EnqSrvResName [%s] invalid resource of Type [SAPNW]",
      $sMyName, $localEnqSrvResName );
      $iReturnCode = 0; # Failure..
    }
    elsif ( $localEnqSrvResName !~ m/\S+/ && $localInstType =~ /^enqrep$/io )  {
      &logWarn ( 20386,
      "<%s> EnqSrvResName is not specified for InstType ENQREP.",
      $sMyName);
    }
  }

  #----------------------------------------------
  # Validate attributes for online and monitor ..
  #----------------------------------------------
  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #---------------------------------------------
    # Validate ProcMon after removing duplicates..
    #--------------------------------------------
    my $rslocalProcMon = \$_[6]; # Get reference to what's passed in..
    my %hsProcMon = map { $_ => 1 } $localProcMon =~ /([^\s]+)/g;
    my @lsProcMon = keys %hsProcMon;
    my ( @lsBadProcs, %hsAllowed ) = ();

    if ( $localInstType =~ /^(appserv)$/io ) {
      if ( ! grep { /^(dw|jc|jstart)$/o } @lsProcMon ) { # Must have..
        &logError ( 20009, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least one of [dw jc jstart] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( dw co se jc jstart ig gw cm cs gwrd icman ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20010, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(enqueue)$/io ) {
      if ( ! grep { /^en/o } @lsProcMon ) {
        &logError ( 20011, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least [en] or [enq] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }

      %hsAllowed = map { $_ => 1 } qw ( en enq ms gw gwrd wd ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20012, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(enqrep)$/io  ) {
      %hsAllowed = map { $_ => 1 } qw ( er enqr ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20013, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif( $localInstType =~ /^(SMDAgent)$/io ) {
      if ( ! grep { /^(jc|jstart)$/o } @lsProcMon ) { # Must have..
        &logError ( 20014, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least one of [dw jc jstart] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(sapstartsrv)$/io  ) {
      if ( ! grep { /^(sapstartsrv)$/o } @lsProcMon ) {
        &logError ( 20013, "<%s> Incorrect value for ProcMon [%s].
        Must be [sapstartsrv] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }
    elsif ( $localInstType =~ /^(hana)$/io  ) {
      if ( ! grep { /^(hdb)$/o } @lsProcMon ) {
        &logError ( 20393, "<%s> Incorrect value for ProcMon [%s].
        Has to include at least [hdb] for InstType [%s]",
        $sMyName, $localProcMon, $localInstType );

        $iReturnCode = 0; # Failure..
      }
      %hsAllowed = map { $_ => 1 } qw ( hdb hdbstatisticsserver hdbpreprocessor hdbcompileserver hdbxsengine sapwebdisp_hdb ); # Restricted set..
      if ( @lsBadProcs = grep { ! $hsAllowed { $_ } } @lsProcMon ) {
        &logError ( 20010, "<%s> Incorrect value(s) [%s] in ProcMon found.
        Must be one of [%s] for InstType [%s]", $sMyName,
        join ( ' ', @lsBadProcs ), join ( ' ', keys %hsAllowed ),
        $localInstType );

        $iReturnCode = 0; # Failure..
      }
    }


    #--------------------
    # Re-create procmon..
    #--------------------
    $$rslocalProcMon = join ( ' ', @lsProcMon );
  }

  #-------------------------------------------
  # Validate attributes for all entry points..
  #-------------------------------------------

  #-----------------------------------------
  # SAPSID.. must be three characters long..
  #-----------------------------------------
  if ( $sSAPSID !~ /^\S{3}$/o ) {
    &logError ( 20015,
    "<%s> SAPSID [%s] has to be exactly 3 non-white alpha-numeric characters long",
    $sMyName, $sSAPSID );
    $iReturnCode = 0; # Failure..
  }

  #-----------------------------------
  # InstName. Last two chars 00 - 99..
  #-----------------------------------
  unless ( $sInstName =~ /(\d{2})$/o ) {
    &logError ( 20016,
    "<%s> Last two characters of InstName [%s] not between 00-99",
    $sMyName, $sInstName );
    $iReturnCode = 0; # Failure..
  }

  #------------------
  # Valid InstTypes..
  #------------------
  my @lsIntTypes = qw ( appserv enqueue enqrep smdagent sapstartsrv hana );

  unless ( grep { /^$localInstType$/io } @lsIntTypes ) {
    &logError ( 20017,
    "<%s> InstType [%s] invalid. Has to be one of [%s]",
    $sMyName, $localInstType, join ( ' ', @lsIntTypes ) );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode ||
  &logWarn ( 20018,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String EnvFile,
                         IN String InstProfile,
                         IN String InstType,
                         IN String ProcMon,
                         IN String SAPAdmin,
                         IN String SAPSID,
                         IN String SAPHome,
                         IN String InstName,
                         IN Integer Level_two,
			 IN Bool   StopService
                         IN Bool   UseSystemD );

 Description : Start logic needed for online entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $iLevel_two                Cycle count to run in-depth monitoring
 $bStopService              Specifies whether SAP Start Service should be stopped or not. 
 $bUseSystemD               When enabled, resource comes Online in system.slice
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sSAPHome,
       $sInstName,
       $iLevel_two,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20019, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  Level_two                [%s]
  StopService              [%s]
  UseSystemD               [%s] ",

  $sMyName, $sEnvFile, $sInstProfile, $sInstType, $sProcMon,
  $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName, $iLevel_two, 
  $bStopService, $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting SAPNW processes..
  #-----------------------------------------
  &logTrace ( 20020,
  "<%s> Killing SAPNW processes for this instance", $sMyName );

  if ( SAPNW::KillSAPNWProcesses ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin,
                                             $sEnvFile,
                                             $bStopService,
                                             $bUseSystemD ) ) {
    &logTrace ( 20021,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20022,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );

  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io )) {
    #--------------------------------------------------------
    # Clean IPC resources for the instance ..
    #--------------------------------------------------------
    if ( $sInstName !~ /^ASCS|^SCS/io ) {
      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";
      ( @lsCmdout, $iExitCode, $iTimeout ) = ();

      &logTrace ( 20023,
      "<%s> Cleaning IPC resources for SAPNW instance with Cmd [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );

      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20024,
        "<%s> Insufficient time to run cleanipc. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );

        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
        SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout ) ) {
        &logTrace ( 20025,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }

      #-------------------
      # Return if failed..
      #-------------------
      $iReturnCode || return $iReturnCode;
    }
    else {
      &logTrace ( 20026,
      "<%s> Skipping cleanipc for InstType [%s].
      Possibility of destroying active lock table", $sMyName, $sInstType );
    }

    my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();

    #------------------
    # Remove kill.sap..
    #------------------
    $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

    if ( -e $sKillFile ) {
      &logTrace ( 20027,
      "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
      push ( @lsFiles, $sKillFile );
    }

    #----------------------
    # Remove shutdown.sap..
    #----------------------
    $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

    if ( -e $sShutdownFile ) {
      &logTrace ( 20028,
      "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
      push ( @lsFiles, $sShutdownFile );
    }

    #-----------------------------
    # Remove collector lock file..
    #-----------------------------
    $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

    if ( -e $sCollLockFile ) {
      &logTrace ( 20029,
      "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
      push ( @lsFiles, $sCollLockFile );
    }

    #--------------------------
    # Remove sender lock file..
    #--------------------------
    $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

    if ( -e $sSendLockFile ) {
      &logTrace ( 20030,
      "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
      push ( @lsFiles, $sSendLockFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20031, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
    }
  }
  #----------------------------------
  # Initiate SAPNW startup log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Starting up [$sResourceName]";
  if ( ! ( $sInstType =~ /(hana|smdagent)/io ) && ( $sInstName !~ /^HDB/io )  ) {
    if ( SAPNW::UpdateSAPNWLog ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20032,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20033,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  if ( $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd() ) {
    if ( ($sEnvFile =~ /\S/ )      &&         # EnvFile is specified 
         ($::gsSystemDEnvFile !~ /\S/)) {        # SystemD-compliant EnvFile is not specified 
      $::gsSystemDEnvFile = $sEnvFile;  
      &logInfo (20457, "<%s> EnvFile[%s] is specified, but SystemD-compliant environment file is not specified in attribute SystemDEnvFile. 
                 The agent will use [%s] as the value for directive EnvironmentFile in the service file.",
                 $sMyName, $sEnvFile, $sEnvFile);   
    } 
  }

  #------------------------------------------------
  # Check if service is already started. 
  # If yes, do not attempt to start it again
  #------------------------------------------------
  my @liDummyPids = ();
  my $iRet = SAPNW::GetProcState ( \@liDummyPids,
                                   $sInstProfile,
                                   "SAPSTARTSRV",
                                   "sapstartsrv",
                                   $sSAPAdmin,
                                   $sEnvFile,
                                   0,              # Use FFLM..
                                   $bStopService,
                                   $bUseSystemD );

  if ( VCSDef_MonitorExit::ONLINE == $iRet ) {
    &logTrace ( 20448,
    "<%s> Found SAPSTARTSRV instance online with Pid[%s]. The agent will not start it again.",
    $sMyName, join (' ',@liDummyPids) );
  }
  else {
    &logTrace ( 20449,
    "<%s> Did not find SAPSTARTSRV instance online with all required processes. Attempting to start it.",
    $sMyName );

    #------------------------------------------------
    # Define sapcontrol command and args variables..
    #------------------------------------------------
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StartService $sSAPSID";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20034,
    "<%s> Starting SAPNW instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20035,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
  
      $iReturnCode = 0; # Failure..
    }
    elsif ($bUseSystemD) {
      &logTrace ( 20398,
        "<%s> SYSTEMD is detected", $sMyName );
      if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartService", undef) ) {
        &logTrace ( 20399,
            "<%s> Start  command [%s] with args [%s] executed successfully",
            $sMyName, $sCommand, $sCommandArgs );
      }
      else {
        &logError ( 20402, "<%s>: Start command [%s] with args [%s] failed to execute",
            $sMyName, $sCommand, $sCommandArgs);
      }
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
      &logTrace ( 20036,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20450,
      "<%s> Cmd [%s] with args [%s] failed to execute.",
      $sMyName, $sCommand, $sCommandArgs );
    }
  
    #-------------------
    # Return if failed..
    #-------------------
    $iReturnCode || return $iReturnCode;
    $iTimeout = $iTillWhen - time ;
    #------------------------------------------------
    # Define sapcontrol command and args variables..
    #------------------------------------------------
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function WaitforServiceStarted $iTimeout 1";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20037,
    "<%s> Waiting for Service to be started for SAPNW instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20423,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
  
    $iReturnCode = 0; # Failure..
    }
    elsif ($bUseSystemD) {
      &logTrace ( 20424,
      "<%s> SYSTEMD is detected", $sMyName );
      if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartServiceWait", undef ) ) {
        &logTrace ( 20425,
            "<%s> Start  command [%s] with args [%s] executed successfully",
            $sMyName, $sCommand, $sCommandArgs );
        if ( $sInstType !~ /^(sapstartsrv)$/io ) {
          sleep (2); # Avoiding the possibility of failure of StartWait command via Systemd
        }
      }
      else {
        &logError ( 20428, "<%s>: Start command [%s] with args [%s] failed to execute",
            $sMyName, $sCommand, $sCommandArgs);
      }
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
      &logTrace ( 20039,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20451,
      "<%s> Cmd [%s] with args [%s] failed to execute.",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  my $iSleepTime = 10; # Secs.. 
  if ( $sInstType !~ /^(sapstartsrv)$/io ) {
    #------------------------------------------
    # Start SAPNW Instance ..
    #------------------------------------------
    my %hsExitCode = ( 0 => 'Success',
                       1 => 'OS configuration error',
                       2 => 'Called with wrong Parameters',
                       3 => 'User not authorized to execute script',
                       4 => 'Environment not set',
                       5 => 'Database startup failed',
                       6 => 'Database on remote DB-Server not running',
                       7 => 'Startup of Instance failed',
                       8 => 'Start profile not found',
                       9 => 'User not authorized to start database' );

    my $sLogFile = (); #"/usr/sap/$sSAPSID/$sInstName/log/" .  "${sSAPSID}_${sInstName}.log"; 
    if ( $sInstType =~ /smdagent/io ) {
      $sLogFile = "/dev/null";
    }
    else {
      $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" . "${sSAPSID}_${sInstName}.log";
    }
    $iTimeout = $iTillWhen - time ;
    #-------------------------------------------
    # Get re-direct string based on user shell..
    #-------------------------------------------
    my $sRdStr = ( getpwnam $sSAPAdmin ) [-1] =~ /csh$/o ? ">>& $sLogFile":
                                                         "1>> $sLogFile 2>&1";
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StartWait $iTimeout 1";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();

    &logTrace ( 20040,
    "<%s> Starting SAPNW instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20035,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ($bUseSystemD) {
      &logTrace ( 20403,
      "<%s> SYSTEMD is detected", $sMyName );
      if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartWait", undef ) ) {
        &logTrace ( 20404,
            "<%s> Start  command [%s] with args [%s] executed successfully",
            $sMyName, $sCommand, $sCommandArgs );
      }
      else {
        &logError ( 20407, "<%s>: Start command [%s] with args [%s] failed to execute",
            $sMyName, $sCommand, $sCommandArgs);
      }
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout ) ) {
      &logTrace ( 20042,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20452,
      "<%s> Cmd [%s] with args [%s] failed to execute.",
      $sMyName, $sCommand, $sCommandArgs );
    }
    #-------------------
    # Return if failed..
    #-------------------
    $iReturnCode || return $iReturnCode;
  }

  #--------------------
  # Delay entry point..
  #--------------------
  $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20043,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPNW::GetProcState ( \@liPids,
                                    $sInstProfile,
                                    $sInstType,
                                    $sProcMon,
                                    $sSAPAdmin,
                                    $sEnvFile,
                                    1,                          # Running FFLM..
                                    $bStopService,
                                    $bUseSystemD ) ) {          
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20044,
      "<%s> Process check confirmed start-up of this instance of SAPNW server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20045,
      "<%s> Process check could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  if ( $sInstType !~ /^(sapstartsrv)$/io && ($iLevel_two >= 1 ) ) {      # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      @liPids = (); # Initialise..

      sleep $iSleepTime; # Sec..

      &logTrace ( 20046,
      "<%s> Confirming start-up with [%s] sec in hand by health check",
      $sMyName, $iTimeout - $iSleepTime );

      #---------
      # Do SLM..
      #---------
      if ( VCSDef_MonitorExit::ONLINE ==                   # Online after..
           SAPNW::SecondLevelMonitor ( $sEnvFile,     # Running SLM..
                                            $sProcMon,
                                            $sInstProfile,
                                            $sInstType,
                                            $sSAPAdmin,
                                            $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20047,
        "<%s> Second Level Monitor confirmed start-up of this instance of SAPNW server",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20048,
        "<%s> Second Level Monitor could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );

        $iReturnCode = 0; # Failure..
      }
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String EnvFile,
                        IN String InstProfile,
                        IN String InstType,
                        IN String ProcMon,
                        IN String SAPAdmin,
                        IN String SAPSID,
                        IN String SAPHome,
                        IN String InstName,
                        IN Bool   StopService,
                        IN Bool   UseSystemD);

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $bStopService              Specifies whether SAP Start Service should be stopped or not. 
 $bUseSystemD               When enabled, resource comes Online in system.slice 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin, 
       $sSAPSID, 
       $sSAPHome, 
       $sInstName,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20049, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  StopService              [%s]
  UseSystemD               [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sInstType, $sProcMon, 
  $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName, $bStopService, 
  $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  my $sSAPWork = "/usr/sap/$sSAPSID/$sInstName/work" ;
  if ( $sInstType !~ /^(sapstartsrv)$/io ) {
    $iTimeout = $iTillWhen - time;
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopWait $iTimeout 1";

    ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

    &logTrace ( 20050,
    "<%s> Stopping SAPNW with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------------
    # Request shutdown..
    #------------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20051,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
            SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout ) ) {
      &logTrace ( 20052,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    } 
    else {
      &logError ( 20053, "<%s> Failed to execute StopWait.", $sMyName );
      $iReturnCode = 0; # Failure..
      my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
      SAPNW::DeletePidFile ( $sResourceName );
      return $iReturnCode ;
    }
  }
  if ( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io )) { #Stop the SAP Start Service
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopService";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20054,
    "<%s> Stopping SAPNW instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request shutdown..
    #------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20055,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin,
                            $sEnvFile,
                            $sCommand,
                            $sCommandArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdout ) ) {
      &logTrace ( 20056,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
      #---------------------------------------------------------------------------------------
      #  StopService call does not block the command until the service is stopped.  
      #  The agent will delay the kill operation to allow the service to be stopped gracefully
      #---------------------------------------------------------------------------------------
      #--------------------
      # Delay entry point..
      #--------------------
      $iSleepTime = 2; # Secs..
      my $iSrvTimeout = 30;
      my $iSrvTillWhen = $iSrvTimeout + time ;
      while ( ( $iSrvTimeout = $iSrvTillWhen - time ) > 5 ) {
        my @liGoners = (); # Initialise..
        sleep $iSleepTime; # Sec..
        &logTrace ( 20445,
        "<%s> Confirming shutdown with [%s] sec in hand",
        $sMyName, $iSrvTimeout - $iSleepTime );
        $iSleepTime = 5; # resetting value to 5 for upcoming iterations
    
        #---------
        # Do FLM..
        #---------
        SAPNW::GetProcState ( \@liGoners,
                              $sInstProfile,
                              $sInstType,
                              $sProcMon,
                              $sSAPAdmin,
                              $sEnvFile,
                              0,
                              $bStopService,
                              $bUseSystemD ) ;

        if ( scalar @liGoners ) {
          &logInfo ( 20446,
          "<%s> Process check could not confirm shutdown. Running processes: [%s], Sleeping for [%s] sec",
          $sMyName, join (' ', @liGoners), $iSleepTime );
        }
        else {
        
          &logTrace ( 20447,
          "<%s> Process check confirmed shutdown for this instance.",
          $sMyName);
          last;
        }      
      }
    }
    else {
      &logError ( 20057, "<%s> Failed to execute StopService.", $sMyName );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------------
  # Whack any persisting SAPNW processes..
  #-----------------------------------------
  &logTrace ( 20058,
  "<%s> Killing SAPNW processes for this instance", $sMyName );

  if ( SAPNW::KillSAPNWProcesses ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin,
                                             $sEnvFile,
                                             $bStopService,
                                             $bUseSystemD ) ) {
    &logTrace ( 20059,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20060,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  if ( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io )) { #Clean up the ipc resources for  the SAP Start Service
    my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();

    #------------------
    # Remove kill.sap..
    #------------------
    $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

    if ( -e $sKillFile ) {
      &logTrace ( 20061,
      "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
      push ( @lsFiles, $sKillFile );
    }

    #----------------------
    # Remove shutdown.sap..
    #----------------------
    $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

    if ( -e $sShutdownFile ) {
      &logTrace ( 20062,
      "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
      push ( @lsFiles, $sShutdownFile );
    }

    #-----------------------------
    # Remove collector lock file..
    #-----------------------------
    $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

    if ( -e $sCollLockFile ) {
      &logTrace ( 20063,
      "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
      push ( @lsFiles, $sCollLockFile );
    }

    #--------------------------
    # Remove sender lock file..
    #--------------------------
    $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

    if ( -e $sSendLockFile ) {
      &logTrace ( 20064,
      "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
      push ( @lsFiles, $sSendLockFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20065, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
      $iReturnCode = 0; # Failure..
    }
  
    #-----------------------------------
    # Check if it's safe to clean IPCs..
    #-----------------------------------
    if ( $sInstName !~ /^ASCS|^SCS/io ) {
      #----------------------------------------
      # Define cleanipc command and arguments..
      #----------------------------------------
      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";
      ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate..

      &logTrace ( 20066,
      "<%s> Stopping SAPNW with Command [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );
  
      #------------------------
      # Request shutdown..
      #------------------------
      $iTimeout = $iTillWhen - time;
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20067,
        "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
  
        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
              SAPNW::ExecLocalCmd ( $sSAPAdmin,
                                         $sEnvFile,
                                         $sCommand,
                                         $sCommandArgs,
                                         \$iExitCode,
                                         $iTimeout,
                                         \@lsCmdout ) ) {
        &logTrace ( 20068,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
    }
    else {
      &logTrace ( 20069,
      "<%s> Skipping cleanipc for InstType [%s].
      Possibility of destroying active lock table", $sMyName, $sInstType );
    }
  }

  #----------------------------------
  # Initiate SAPNW shutdown log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Shutdown completed for [$sResourceName]";
  if ( ! ( $sInstType =~ /(hana|smdagent)/io )  && ( $sInstName !~ /^HDB/io )  ) {
    if ( SAPNW::UpdateSAPNWLog ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20070,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20071,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  SAPNW::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String EnvFile,
                        IN String InstProfile,
                        IN String InstType,
                        IN String ProcMon,
                        IN String SAPAdmin, 
                        IN String SAPSID, 
                        IN String SAPHome, 
                        IN String InstName 
                        IN Bool   StopService
                        IN Bool   UseSystemD );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sSAPSID                   SAP System ID
 $sSAPHome                  SAP Instance Home
 $sInstName                 SAP Instance Name
 $bStopService              Specifies whether SAP Start Service should be stopped or not. 
 $bUseSystemD               When enabled, resource comes Online in system.slice
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sEnvFile,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sSAPSID,
       $sSAPHome,
       $sInstName,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20072, "Subroutine <%s> called with:
  CleanReason              [%s]
  EnvFile                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  SAPHome                  [%s]
  InstName                 [%s]
  StopService              [%s]
  UseSystemD               [%s]",
 
  $sMyName,  $iCleanReason,  $sEnvFile,  $sInstProfile, $sInstType,
  $sProcMon, $sSAPAdmin, $sSAPSID, $sSAPHome, $sInstName ,$bStopService, 
  $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsCmdout, @liDummyPids, @liPids ) = ();

  my ($sKillFile, $sShutdownFile, $sCollLockFile, $sSendLockFile, @lsFiles ) = ();

  if ( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io )) { #Stop the SAP Start Service

    my $iInstID = ();
    ($iInstID ) = $sInstName =~ /(\d{2})$/o;
    $sCommand = "$sSAPHome/sapcontrol";
    $sCommandArgs = "-nr $iInstID -function StopService";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20073,
    "<%s> Stopping SAPNW instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request shutdown..
    #------------------
    $iTimeout = $iTillWhen - time;
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20074,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin,
                            $sEnvFile,
                            $sCommand,
                            $sCommandArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdout ) ) {
      &logTrace ( 20075,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
      sleep (2); # sec... This is to ensure graceful shutdown
    }
    else {
      &logWarn ( 20387, "<%s> Failed to execute StopService. Run the command <%s %s> manually to stop sapstartsrv process if exists on the system.", $sMyName, $sCommand, $sCommandArgs);
      $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
    }
  
  }
  
  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #----------------------------
    # Get SAP PIDs
    #----------------------------
    SAPNW::GetProcState ( \@liDummyPids,
                           $sInstProfile,
                           $sInstType,
                           $sProcMon,
                           $sSAPAdmin,
                           $sEnvFile,
                           0,              # Use RFLM..
                           $bStopService,
                           $bUseSystemD );

    @liPids = $bSapStart ? shift @liDummyPids: @liDummyPids; # Local..

    if( $bSapStart && defined $iCCMSX ) {
      push ( @liPids, $iCCMSX );
    }

    if( $bSapStart && defined $iCCMSR ) {
      push ( @liPids, $iCCMSR );
    }

    #-----------------------------------------------
    # Attempt to request sapstart to stop instance..
    #-----------------------------------------------
    if ( scalar @liPids ) {
      &logInfo ( 20077, "<%s> Sending SIGINT to PIDs [%s]",
      $sMyName, join ( ' ', @liPids ) );

      #-----------------------------------------
      # Give SAP chance to shutdown gracefully..
      #-----------------------------------------
      if ( kill ( 2, @liPids ) == scalar @liPids ) {
        &logInfo ( 20078,
        "<%s> Successfully sent SIGINT to PIDs [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
      else {
        &logWarn ( 20079,
        "<%s> Failed to send SIGINT to PID [%s]",
        $sMyName, join ( ' ', @liPids ) );
      }
    }
    else {
      &logError ( 20080,
      "<%s> Proceeding with SIGKILL for SAP Instance [%s]",
      $sMyName, $sInstName );
    }
  } 

  #-----------------------------------------
  # Whack any persisting SAPNW processes..
  #-----------------------------------------
  &logTrace ( 20081,
  "<%s> Killing SAPNW processes for this instance", $sMyName );

  if ( SAPNW::KillSAPNWProcesses ( $sInstProfile,
                                             $sInstType,
                                             $sProcMon,
                                             $sSAPAdmin ,
                                             $sEnvFile,
                                             $bStopService,
                                             $bUseSystemD ) ) {
    &logTrace ( 20082,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20083,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  my $iInstID = ();
  #-----------------------------------
  # Check if it's safe to clean IPCs..
  #-----------------------------------
  if ( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io )) { 
    if ( $sInstName !~ /^ASCS|^SCS/io ) {
      ($iInstID ) = $sInstName =~ /(\d{2})$/o;

      $sCommand = "$sSAPHome/cleanipc";
      $sCommandArgs = "$iInstID remove";

      #------------------------
      # Run cleanipc ..
      #------------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20084,
        "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );

        $iReturnCode = 0; # Failure..
      }
      elsif ( SAPNW::ExecLocalCmd ( $sSAPAdmin,
                                         $sEnvFile,
                                         $sCommand,
                                         $sCommandArgs,
                                         \$iExitCode,
                                         $iTimeout,
                                         \@lsCmdout ) ) {
        &logTrace ( 20085,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
      }
      else {
        &logWarn ( 20388, "<%s> Failed to execute cleanipc. Run the command <%s %s> manually.", $sMyName, $sCommand, $sCommandArgs);
        $iReturnCode = 1; # Failure,still forcefully succeed the clean operation
      }
    }
    else {
      &logTrace ( 20087,
      "<%s> Skipping cleanipc for InstType [%s].", $sMyName, $sInstType );
    }

    #------------------
    # Remove kill.sap..
    #------------------
    $sKillFile = "/usr/sap/$sSAPSID/$sInstName/work/kill.sap";

    if ( -e $sKillFile ) {
      &logTrace ( 20088,
      "<%s> Located file [%s] for removal", $sMyName, $sKillFile );
      push ( @lsFiles, $sKillFile );
    }

    #----------------------
    # Remove shutdown.sap..
    #----------------------
    $sShutdownFile = "/usr/sap/$sSAPSID/$sInstName/work/shutdown.sap";

    if ( -e $sShutdownFile ) {
      &logTrace ( 20089,
      "<%s> Located file [%s] for removal", $sMyName, $sShutdownFile );
      push ( @lsFiles, $sShutdownFile );
    }

    #-----------------------------
    # Remove collector lock file..
    #-----------------------------
    $sCollLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgcpid";

    if ( -e $sCollLockFile ) {
      &logTrace ( 20090,
      "<%s> Located file [%s] for removal", $sMyName, $sCollLockFile );
      push ( @lsFiles, $sCollLockFile );
    }

    #--------------------------
    # Remove sender lock file..
    #--------------------------
    $sSendLockFile = "/usr/sap/$sSAPSID/$sInstName/data/rslgspid";

    if ( -e $sSendLockFile ) {
      &logTrace ( 20091,
      "<%s> Located file [%s] for removal", $sMyName, $sSendLockFile );
      push ( @lsFiles, $sSendLockFile );
    }

    #-------------------
    # Remove the files..
    #-------------------
    if ( scalar @lsFiles && ! Sys::UnlinkFilesAsUser ( @lsFiles ) ) {
      &logError ( 20092, "<%s> Failed to remove stale file(s) [%s]",
      $sMyName, join ( ' ', @lsFiles ) );
      $iReturnCode = 0; # Failure..
    }
  }

  #----------------------------------
  # Initiate SAPNW cleanup log..
  #----------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sMessage = "Clean called for [$sResourceName] [$sCleanReason]";

  if ( ! ( $sInstType =~ /(hana|smdagent)/io )  && ( $sInstName !~ /^HDB/io )  ) {
    if ( SAPNW::UpdateSAPNWLog ( $sInstName,
                                           $sInstProfile,
                                           $sInstType,
                                           $sSAPAdmin,
                                           $sSAPSID,
                                           $sMessage ) ) {
      &logTrace ( 20093,
      "<%s> Updated log successfully", $sMyName );
    }
    else {
      &logError ( 20094,
      "<%s> Could not log successfully", $sMyName );
    }
  }

  SAPNW::DeletePidFile ( $sResourceName );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String InstProfile,
                                   IN String InstType,
                                   IN String ProcMon,
                                   IN String SAPAdmin, 
                                   IN String EnvFile,
                                   IN Bool   StopService,
                                   IN Bool   UseSystemD );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $bStopService              Specifies whether SAP Start Service should be stopped or not.
 $bUseSystemD               When enabled, resource comes Online in system.slice
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20095, "Subroutine <%s> called:
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  StopService              [%s]
  UseSystemD               [%s]",

  $sMyName, $sInstProfile, $sInstType, $sProcMon, $sSAPAdmin ,  
  $sEnvFile, $bStopService, $bUseSystemD );
  

  my @liDummyPids = ();
  my $iReturnCode = SAPNW::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               $sProcMon,
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               1,              # Use FFLM..
                                               $bStopService,
                                               $bUseSystemD );

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20096,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20097,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }
                                                                                
  &logTrace ( 20098,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String InstProfile,
                               IN String InstType,
                               IN String ProcMon,
                               IN String SAPAdmin,
                               IN String EnvFile,
                               IN Integer UseFFLM, 
                               IN Bool StopService,
                               IN Bool UseSystemD );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $iUseFFLM                  Filtering procedure
 $bStopService              Specifies whether SAP Start Service should be stopped or not.
 $bUseSystemD               When enabled, resource comes Online in system.slice
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $iUseFFLM,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20099, "Subroutine <%s> called with:
  OnlinePids               [%s]
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  UseFFLM                  [%s]
  StopService              [%s]
  UseSystemD               [%s]",

  $sMyName, $rliOnlinePids, $sInstProfile, $sInstType, $sProcMon,
  $sSAPAdmin, $sEnvFile, $iUseFFLM, $bStopService, $bUseSystemD );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp, $iSapstartPid,
       @hProcListHash, @lsUpdates, @lsValues, $iDwMaster, $iJcFlag,
       %hiProcPids, %hsSrvrDetails, %hsCopy, $sEntryPointName,
       $iSleepTime, $sInstName, $sSAPSID, @liDummyPids, %hsProcMonCopy, 
       $iJstartFlag, $iJstartmFlag, $sSAPHost, @lsWords, @lsOthers,$iErPid ,
       $bNWVersionLt71, $iHdbFlag) = ();

  my ( $sSwitch, $bSapstartFlag, $bSapstartSrvFlag,$bErFlag,$iPosn, $bAddIn, %hiDwPids, $iParent, $i ) = ();
  $iParent = 1 ;  # init
  $bNWVersionLt71 = 0 ;
  $iPosn = 0;

  @$rliOnlinePids = ();
  
  #--------------------------
  # Other processes to kill..
  #--------------------------
  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);

  my $sStartProfile = "/usr/sap/$sSAPSID/SYS/profile/START_${sInstName}_${sSAPHost}";
  
  if( -f $sStartProfile ) {
    $bNWVersionLt71 = 1 ;
    &logTrace ( 20100,
    "<%s> SAP Start Profile [%s] exists.",
    $sMyName, $sStartProfile );
  }

  @lsOthers = ( "pf=.*profile/(START|${sSAPSID})_${sInstName}_${sSAPHost}", "hdbstatisticsserver", "hdbpreprocessor", "hdbcompileserver", "hdbxsengine", "sapwebdisp_hdb" );

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if ( $bNWVersionLt71 && $hsProcs { jc } && $hsProcs { dw } ) {
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $bNWVersionLt71 && $hsProcs { JC }  ) { # Recursive call..
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $hsProcs { jc } ) {
    $sArch = 1;
  }
  elsif ( $hsProcs { jstart } && $hsProcs { dw } ) {
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $hsProcs { JSTARTM } ) { # Recursive call..
    $sArch = 3;
    ++ $bAddIn;
  }
  elsif ( $hsProcs { dw } ) {
    $sArch = 2;
  }

  if ( $sArch ) {
    &logTrace ( 20101,
    "<%s> SAP Instance Architecture is [%s]",
    $sMyName, $hsArch { $sArch } );
  }
  else {
    &logTrace ( 20102,
    "<%s> SAP Instance Type is [%s]",
    $sMyName, uc $sInstType );
  }

  #-----------------------------------------------
  # Insert sapstart, unless its a recursive call..
  #-----------------------------------------------
  if ( $sProcMon eq 'JSTART' ) { # It's Java ..
    $sProc = 'jstart';
  }
  elsif ( $sProcMon eq 'JSTARTM' ) { # It's AddIn ..
    $sProc = 'dw jstart'; # AddIn for recursive call..
  }
  elsif ( $sProcMon eq 'sapstartsrv' ) { # It's SAPSTARTSRV ..
    $sProc = 'sapstartsrv'; 
  }
  elsif ( $sInstType =~ /^(hana)$/io ) {
    $sProc = "sapstart sapstartsrv hdbnameserver hdbindexserver $sProcMon"; 
  }
  elsif ( $bNWVersionLt71 && ( $sProcMon eq 'jlaunch' ) ) {
    $sProc = $sProcMon;  # AddIn for recursive call..
  }
  elsif ( $bNWVersionLt71 && ( $sProcMon eq 'JC' ) ) {
    $sProc = 'jc dw'; # ADD-IN for recursive call..
  }
  else {
    $sProc =  "sapstart sapstartsrv $sProcMon";
  }

  $iUid = getpwnam $sSAPAdmin;

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == $iUid ) &&      # User is the super-user..
       ( 2 == $iUseFFLM ) ) { # And called to kill-by-UID..
    &logTrace ( 20103, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sSAPAdmin, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPNW::GetSrvrDetails ( $sInstProfile, 
                                      $sProc, 
                                      $bNWVersionLt71,
                                      \%hsSrvrDetails ) ) {

      delete $hsSrvrDetails { dw } if $sProcMon eq 'JSTARTM'; # Remove dw in recursive..
      delete $hsSrvrDetails { dw } if ( $bNWVersionLt71 && ($sProcMon eq 'JC')); # Remove dw in recursive..

      &logTrace ( 20104,
      "<%s> Got details for monitoring SAPNW instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20105,
      "<%s> Failed to get details for monitoring SAPNW instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $sEntryPointName ne 'offline' && $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPNW process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPNW::FFLMStatus ( $sPidFile,
                                             $sProc,
                                             $iUid,
                                             \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          if ( $sKey eq 'sapstart' ) { # Check sapstart found ..
            $iSapstartPid = $iPid; 
            unshift ( @$rliOnlinePids, $iPid ); # Insert at the front..
          }
          elsif ( $sKey eq 'sapstartsrv' ) { # Check sapstartsrv found ..
            $iSapStartSrv = $iPid; 
            ++ $bSapstartSrvFlag; # Mark it found..
            ++ $bSapStartSrv;     # Mark global..
          }
          else {
            if ( $sKey eq 'dw' ) {  # Check dw master found for ABAP or AddIn ..
              $iDwMaster = $iPid; 
            }
            elsif ( $sKey eq 'jc' ) { # Check jc found for Java ..
              ++ $iJcFlag;
            }
            elsif ( $sKey eq 'jstartm' ) { # Check jstart master found for AddIn ..
              ++ $iJstartmFlag;
            }
            elsif ( $sKey eq 'jstart' ) {
              ++ $iJstartFlag;
            }
            elsif ( $sKey eq 'er' ) {
              ++ $bErFlag;
              $iErPid = $iPid; 
            }
            elsif ( $sKey eq 'hdb' ) {
              ++ $iHdbFlag;
            }

            &logTrace ( 20106, "<%s> [%s] found in FFLM [%s]", $sMyName, $sKey, $iPid);

            push ( @$rliOnlinePids, $iPid ); # Push..
          }
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20107,
      "<%s> SAPNW instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20108,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sSAPAdmin, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20109,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20110,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20111,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    my $iParent = 1; # Set to init..
    my $sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '';

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20112, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID }, $$rhlProc { COMMAND } );

      #--------------------------------
      # Handle Solaris 10 local zones..
      #--------------------------------
      if ( $sOSVer eq '5.10'                       && # Solaris 10..
           $$rhlProc { COMMAND } =~ m@/sbin/init$@ && # With init..
           ! kill ( 0, 1 ) ) {                        # In a non-global zone..
        $iParent = $$rhlProc { PPID };                # Get actual PID..
      }

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  #-----------------------
                  # Handle dw separately..
                  #-----------------------
                  $iFound = VCSDef_MonitorExit::ONLINE unless $sKey eq 'dw';

                  #-----------------------
                  # Handle special cases..
                  #-----------------------
                  $sSwitch = $sKey =~ /^(jc|jstart|dw|sapstart|sapstartsrv|jlaunch|hdb)$/o ? $sKey: 'default';

                  if("sapstart" eq $sSwitch) {
                      $iPosn = 0;
                      $iSapstartPid = $iPid; # Store..
                  }

                  elsif("jc" eq $sSwitch) {
                      $iPosn = 1;
                      ++ $iJcFlag; # Set flag..
                  }
                  elsif("jstartm" eq $sSwitch) {
                      $iPosn = 2;
                      ++ $iJstartmFlag;
                  }

                  elsif("jstart" eq $sSwitch) {
                      $iPosn = 3;
                      ++ $iJstartFlag;
                  }
                  elsif("dw" eq $sSwitch) {
                      $iPosn = 4;
                      $hiDwPids { $iPid } = $$rhlProc { PPID }; # Store..
                  }
                  elsif("jlaunch" eq $sSwitch) {
                      $iPosn = 5;
                  }
                  elsif("sapstartsrv" eq $sSwitch) {
                      $iSapStartSrv = $iPid; # Store..
                      ++ $bSapstartSrvFlag; # Mark it found..
                      ++ $bSapStartSrv;     # Mark global..
		                  if ( $sInstType !~ /^(sapstartsrv)$/io ) {
                        $hiProcPids { $sKey } = $iPid;                # Store..
                        push ( @lsUpdates, $sPidFile, $sProc, $iFound );
                        if ( !$bStopService) {
                          next;
                        }
                    }
                  }
                  elsif("er" eq $sSwitch) {
                      ++ $bErFlag;
                      $iErPid = $iPid;
                  }
                  elsif("hdb" eq $sSwitch) {
                      ++ $iHdbFlag;
                  }
 
                  else{ 
                      $iPosn = scalar @$rliOnlinePids;
                  }

                  #--------------------------------
                  # Insert into appropriate place..
                  #--------------------------------
                  $iPosn = scalar @$rliOnlinePids if $iPosn > scalar @$rliOnlinePids;

                  splice ( @$rliOnlinePids, $iPosn, 0, $iPid ); # Insert..
                  $hiProcPids { $sKey } = $iPid;                # Store..
    
                  &logTrace ( 20113,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc );
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        elsif ( $sProcMon !~ /JSTART/ ) { # Request for kill..
          &logTrace ( 20114,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( $sInstType !~ /^(sapstartsrv)$/io )    &&         # Instance Type is not SAPSTARTSRV
	   ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&         # Not caught earlier..
           $$rhlProc { UID } == $iUid               &&         # Owned by SAPAdmin
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..

        #-----------------------------------------
        # Populate dwindling sapstartsrv process..
        #-----------------------------------------
	      if ( $$rhlProc { COMMAND } =~ /sapstartsrv/o ) {
	        $iSapStartSrv = $iPid; # Populate global..
          if ( !$bStopService) {
            next;
          }
	      }
        elsif ( $$rhlProc { COMMAND } =~ /cm\.sap/o ) { # Note sapccm4x process
          $iCCMSX = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /cs\.sap/o ) { # Note sapccmsr process
          $iCCMSR = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /sapccm4x/o ) {
          $iCCMSX = $iPid; # Populate global..
        }
        elsif ( $$rhlProc { COMMAND } =~ /sapccmsr/o ) { # Note sapccmsr process
          $iCCMSR = $iPid; # Populate global..
        }

        push ( @$rliOnlinePids, $iPid );                       # Store..
        &logTrace ( 20115,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline ) = ();

  my @liDwPids = keys %hiDwPids;
  $sTemp = ''; # Ensure..

  foreach $sKey ( sort keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) = splice ( @lsValues, 0, 3 ) ) {
      #-------------------------------------
      # Set status for dw master process ..
      #-------------------------------------
      if ( $bAddIn && $sProcMon !~ /JSTARTM/ ) {
        &logTrace ( 20116, 
        "<%s> Key [%s] EntryPoint [%s] AddIn [%s] dw master [%s]",
        $sMyName, $sKey, $sEntryPointName, $bAddIn, $iDwMaster );
      }

      if ( $sKey eq 'dw' && scalar @liDwPids ) {
        #-------------------------------------
        # Check if we the dw master process ..
        #-------------------------------------
        if ( $iSapstartPid &&                    # sapstart exists
             ( ( $iDwMaster ) =                  # Get the dw master 
             grep {                              # By finding
               $iSapstartPid == $hiDwPids { $_ } # sapstart as it's parent
             } @liDwPids ) ) {                   # From all dw processes
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logTrace ( 20117,
          "<%s> Found master dw process PID [%s] PPID [%s]",
          $sMyName, $iDwMaster, $iSapstartPid );
        }
        elsif ( ( ( $iDwMaster ) =               # A dw process..
                grep {                           # Found..
                  $iParent == $hiDwPids { $_ } &&# With init as parent..
                  ++ $i                          # Mark it..
                } @liDwPids ) && 1 == $i ) {     # If only one such exists..
          $hiProcPids { $sKey } = $iDwMaster;    # Store..
          $iFound = VCSDef_MonitorExit::ONLINE;  # And mark online..
          &logWarn ( 20118,
          "<%s> Missing sapstart process.
          Found master dw process PID [%s] PPID [%s]",
          $sMyName, $iDwMaster, $hiDwPids { $iDwMaster } );
        }
        else {
          &logTrace ( 20119,
          "<%s> Missing dw master. Marking status as offline", $sMyName );

          $iFound = VCSDef_MonitorExit::OFFLINE; # Ensure its offline..

          #------------------------
          # Delete PID, if stored..
          #------------------------
          if ( defined $hiProcPids { $sKey } ) {
            delete $hiProcPids { $sKey };
          }
        }
      }

      #--------------------------
      # Mark status of sapstart..
      #--------------------------
      if ( $sKey eq 'sapstart' && $iSapstartPid ) {
        ++ $bSapstartFlag; # Mark it found..
        ++ $bSapStart;     # Mark global..
      }
      if ( $sKey eq 'sapstartsrv' && $iSapStartSrv ) {
        ++ $bSapstartSrvFlag; # Mark it found..
        ++ $bSapStartSrv;     # Mark global..
      }
      if ( $sKey eq 'er' && $iErPid ) {
        ++ $bErFlag;
      }

      #------------------------------------
      # If we are AddIn, validate jstart ..
      #------------------------------------
      if ( $sKey eq 'jstartm'            && # Looking for jstart..
           $sEntryPointName eq 'monitor' && # in monitor..
           $bAddIn                       && 
           $iDwMaster                   ) { # dw and jstart declared..

        #-----------------------------------
        # Calculate sleep time dynamically..
        #-----------------------------------
        my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
        $iSleepTime = ( $iTillWhen - time ) / 10;
        my $iTimeRemain = undef;

        while ( ! ( $iJstartmFlag && $iJstartFlag ) ) { # jstart master missing..
           
          unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
            &logTrace ( 20120,
            "<%s> Insufficient time [%s] sec to retry. Exiting ..", $sMyName, $iTimeRemain );

            last;
          }

          &logWarn ( 20121,
          "<%s> Missing 'jstart' master, though 'dw' master [%s] exists.
          Retrying in [%s] sec", $sMyName, $iDwMaster, $iSleepTime );

          sleep $iSleepTime; # Take a break..

          #-------------------------------
          # Call ourselves with 'JSTARTM'..
          #-------------------------------
          $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               'JSTARTM',      # In CAPS..
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               0,               # Regular FLM..
                                               $bStopService,
                                               $bUseSystemD );           

          #--------------------------------------------------------------------
          # If we returned online, we know we found jstart master and at least 
          # one jstart server process. PID now is in first position..
          #--------------------------------------------------------------------
          if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
            &logTrace ( 20122,
            "<%s> Found jstart master process after recursive call",
            $sMyName );

            #-----------------------------
            # Update jstartm and jstart ..
            #-----------------------------
            ( $hiProcPids { jstartm }, $hiProcPids { jstart } ) = (); # Ensure undefs..so that PID 
                                                                      # files are not created again for FFLM..

            ++ $iJstartmFlag; # Set flag..
            ${ $hsCopy { jstartm } } [-1] = $iFound; # Mark online..

            ++ $iJstartFlag; # Set flag..
            ${ $hsCopy { jstart } } [-1] = $iFound; # Mark online..
          }
          else {
            &logError ( 20123, "<%s> Missing jstart child process after recursive call. Retrying ..", $sMyName );

            ${ $hsCopy { jstartm } } [-1] = VCSDef_MonitorExit::ONLINE;
            ${ $hsCopy { jstart } } [-1] = VCSDef_MonitorExit::OFFLINE;

            $iFound = VCSDef_MonitorExit::OFFLINE;
          }
        }
      }

      #--------------------------------------------------
      # Check if 'jstart' is missing and 'jc' is alive ..
      #--------------------------------------------------
      if ( $sEntryPointName eq 'monitor' && # In monitor..
           $sKey eq 'jstart'             && # Jstart,
           $iJcFlag                     ) { # when jc, alive..

        #-----------------------------------
        # Calculate sleep time dynamically..
        #-----------------------------------
        my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
        $iSleepTime = ( $iTillWhen - time ) / 10;

        my $iTimeRemain = undef;

        while ( ! $iJstartFlag ) { # jstart missing..

          unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
            &logTrace ( 20124,
            "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );

            last;
          }

          &logWarn ( 20125,
          "<%s> Missing process [%s], though 'jc' exists. Retrying in [%s] sec",
          $sMyName, $sKey, $iSleepTime );

          sleep $iSleepTime; # Take a break..

          #--------------------------------------
          # Call ourselves again with 'jstart' ..
          #--------------------------------------
          $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               'JSTART',
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               0,              # Regular FLM..
                                               $bStopService,
                                               $bUseSystemD );
          #--------
          # Store..
          #--------
          if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
            $hiProcPids { jstart } = (); 
            ${ $hsCopy { jstart } } [-1] = $iFound; # Mark online..
            ++ $iJstartFlag;

            &logTrace ( 20126,
            "<%s> Found jstart process after recursive call.", $sMyName );
          }
          else {
            &logError ( 20127, "<%s> Missing jstart process after recursive call. Retrying ..", $sMyName );
            $iFound = VCSDef_MonitorExit::OFFLINE; # Ignore partial..
          }
        }
      }

      #----------------------------------
      # Check for jc and jlaunch processes 
      # if  $bNWVersionLt71 is set
      #--------------------------------
      
      
      #--------------------------------
      # If we are ADD-IN, validate jc..
      #--------------------------------
      if ( $bNWVersionLt71               && # SAP Netweaver version is less than 7.1 
           $sKey eq 'jc'                 && # Looking for jc..
           $sEntryPointName eq 'monitor' && # in monitor..
           $bAddIn                       && # dw and jc declared..
           $iDwMaster ) {                   # dw master exists..
        if ( ! $iJcFlag ) {                 # jc missing..
           #-----------------------------------
           # Calculate sleep time dynamically..
           #-----------------------------------
           my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
           $iSleepTime = ( $iTillWhen - time ) / 10;
           my $iTimeRemain = undef;
	    
           while ( ! $iJcFlag ) { # jc missing..

             unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
               &logTrace ( 20128,
               "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );

               last;
             }
            &logWarn ( 20129,
            "<%s> Missing process [%s], though 'dw' master [%s] exists.
            Retrying in [%s] sec", $sMyName, $sKey, $iDwMaster, $iSleepTime );

            sleep $iSleepTime; # Take a break..

            #---------------------------
            # Call ourselves with 'JC'..
            #---------------------------
            $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                               $sInstProfile,
                                               $sInstType,
                                               'JC',
                                               $sSAPAdmin,
                                               $sEnvFile,
                                               0,              # Regular FLM..
                                               $bStopService,
                                               $bUseSystemD );

            #--------------------------------------------------------------------
            # If we returned online, we know we found jc and at least one jaunch.
            # PID of jc now is in first position..
            #--------------------------------------------------------------------
            if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
              &logTrace ( 20130,
              "<%s> Found jc process with PID [%s] after recursive call",
              $sMyName, $liDummyPids [0] );

              $hiProcPids { $sKey } = shift @liDummyPids; # Store..
              ++ $iJcFlag; # Set flag..

              #-----------------
              # Update jlaunch..
              #-----------------
              $hiProcPids { jlaunch } = shift @liDummyPids;   # Store..
              ${ $hsCopy { jlaunch } } [-1] = $iFound; # Mark online..
            }
            else {
              &logWarn ( 20131, "<%s> Missing jc after recursive call.Retrying..", $sMyName );
            }
          }
        }
     }

     #--------------------------------------------------
     # Check if 'jlaunch' is missing and 'jc' is alive..
     #--------------------------------------------------
     if ( $bNWVersionLt71               &&           # SAP Netweaver version is less than 7.1
          $sEntryPointName eq 'monitor' &&           # In monitor..
          $sKey eq 'jlaunch'            &&           # Jaunch,
          $iJcFlag                      &&           # when jc, alive..
          VCSDef_MonitorExit::OFFLINE == $iFound ) { # is not..
       #-----------------------------------
       # Calculate sleep time dynamically..
       #-----------------------------------
       my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
       $iSleepTime = ( $iTillWhen - time ) / 10;
       my $iTimeRemain = undef;
       my $iJLaunchFlag = 0 ;
       while ( ! $iJLaunchFlag ) { # jlaunch  missing..

         unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
           &logTrace ( 20132,
           "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );

           last;
         }
	   
         &logWarn ( 20133,
	       "<%s> Missing process [%s], though 'jc' exists. Retrying in [%s] sec",
      	 $sMyName, $sKey, $iSleepTime );
	   
	       sleep $iSleepTime; # Take a break..
	     
      	 #--------------------------------------
	       # Call ourselves again with 'jlaunch'..
	       #--------------------------------------
         $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                            $sInstProfile,
                                            $sInstType,
                                            'jlaunch',
                                            $sSAPAdmin,
                                            $sEnvFile,
                                            0,              # Regular FLM
                                            $bStopService,
                                            $bUseSystemD );
	 #--------
      	 # Store..
	 #--------
         if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
	         if ( scalar @liDummyPids ) {
	           push ( @$rliOnlinePids, @liDummyPids );
	           $hiProcPids { $sKey } = $liDummyPids [0];
	           ++ $iJLaunchFlag ; 
	           &logTrace ( 20134,
	           "<%s> Found jlaunch process with PID [%s] after another attempt",
             $sMyName, $liDummyPids [0] );
	         }
           else {
             &logWarn ( 20135, "<%s> Missing jlaunch after recursive call.Retrying..", $sMyName );
           }
         }
       }
     }
       

     #-----------------------------------------------------------------------
     # Check if hdb.sap<SID>_<INSTANCE> is alive,other hana process is missing
     #-----------------------------------------------------------------------
     if ( ( $sEntryPointName =~ /^(monitor|imf_register)$/o )                  &&           # In monitor or imf_register..
          ( $sKey =~ /^(hdbnameserver|hdbindexserver|hdbstatisticsserver|
                        hdbpreprocessor|hdbcompileserver|
                        hdbxsengine|sapwebdisp_hdb)$/io )                  &&           # hana process other than sapstart|sapstartsrv|hdb.sap<SID>_<INSTANCE>,
          $iHdbFlag                                                        &&           # when hdb.sap<SID>_<INSTANCE>, alive..
          VCSDef_MonitorExit::OFFLINE == $iFound ) { # is not..
       #-----------------------------------
       # Calculate sleep time dynamically..
       #-----------------------------------
       my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
       $iSleepTime = ( $iTillWhen - time ) / 10;
       my $iTimeRemain = undef;
       my $iHanaProcessesFound = 0 ;
       while ( ! $iHanaProcessesFound ) { # hana process  missing..

         unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
           &logTrace ( 20394,
           "<%s> Insufficient time [%s] sec to retry. Exiting..", $sMyName, $iTimeRemain );

           last;
         }
	   
         &logWarn ( 20395,
	       "<%s> Missing process [%s], though 'hdb.sap' exists. Retrying in [%s] sec",
      	 $sMyName, $sKey, $iSleepTime );
	   
	       sleep $iSleepTime; # Take a break..
	     
      	 #--------------------------------------
	       # Call ourselves again with same ProcMon list.
	       #--------------------------------------
         $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                            $sInstProfile,
                                            $sInstType,
                                            $sProcMon,
                                            $sSAPAdmin,
                                            $sEnvFile,
                                            0,              # Regular FLM..
                                            $bStopService,
                                            $bUseSystemD ); 
	       #--------
      	 # Store..
	       #--------
         if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
	         if ( scalar @liDummyPids ) {
	           push ( @$rliOnlinePids, @liDummyPids );
	           $hiProcPids { $sKey } = $liDummyPids [0];
	           ++ $iHanaProcessesFound ; 
	           &logTrace ( 20396,
	           "<%s> Found [%s] process with PID [%s] after another attempt",
             $sMyName, $sKey , $liDummyPids [0] );
	         }
           else {
             &logWarn ( 20397, "<%s> Missing [%s] after recursive call.Retrying..", $sMyName, $sKey );
           }
         }
       }
     }

      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM && $iUseFFLM != 2               &&      # In RFLM..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20136,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPNW::CreatePidFile ( $sPidFile, $hiProcPids { $sKey } ) ) {
          &logTrace ( 20137,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20138,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        if( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io ) ) {
          $iOnline = $iFound;
        }
        elsif( $sKey ne 'sapstartsrv' ) {
          $iOnline = $iFound;
        }
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound && $sKey ne 'sapstart' ) {
        if ( $bAddIn && $sKey eq 'jstart' && $sProcMon !~ /JSTART/ ) {
          $iFound = 'XXX';
        }
        else {
          if( ( $bStopService ) || ( $sInstType =~ /^(sapstartsrv)$/io ) ) {
            $iOffline = $iFound;
	        }
          elsif( $sKey ne 'sapstartsrv' ) {
            $iOffline = $iFound;
          }
        }
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOnline && $iOffline ? VCSDef_MonitorExit::PARTIAL_ONLINE:
                 $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #$sTemp =~  s/XXX/VCSDef_MonitorExit::ONLINE == $iReturnCode ? 110:100/e;

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    $sTemp =~ s/XXX/110/;
  }
  else {
    $sTemp =~ s/XXX/100/;
  }

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20139,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPNW.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #---------------------
  # Report on sapstart..
  #---------------------
  if ( ! $bSapstartFlag                               &&          # Missing sapstart..
       $sEntryPointName eq 'monitor'                  &&          # In monitor..
       $sProcMon !~ /^(JSTART|JSTARTM|JC|jlaunch|sapstartsrv)$/o  &&          # Not in recursive call..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

    &logTrace ( 20140,
    "<%s> Missing sapstart process in first level monitor",
    $sMyName );
  }

  #---------------------
  # Report on sapstartsrv..
  #---------------------
  if ( ! $bSapstartSrvFlag                            &&          # Missing sapstartsrv..
       $bSapstartFlag                                 &&          #sapstart is present
       ( $sInstType !~ /^(sapstartsrv)$/io )          &&
       $sEntryPointName eq 'monitor'                  &&          # In monitor..
       $sProcMon !~ /^(JSTART|JSTARTM|JC|jlaunch)$/o  &&          # Not in recursive call..
       VCSDef_MonitorExit::OFFLINE != $iReturnCode ) { # Not offline..

     my $sHaresCmd = $sCmdHares;
     my $sHagrpCmd = $sCmdHagrp;
     my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
     my $iExitCode = () ;
     #----------------
     # Re-initialise..
     #----------------
     my $sHaresCmdArgs = "-value $sResourceName Group";
  
     #---------------------------------------------
     # Assume the group name is valid and proceed..
     #---------------------------------------------
     my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
     #-----------------------------------------
     # Check if command executed successfully..
     #-----------------------------------------
     if ( $iExitCode = $? >> 8 ) {
        &logError ( 20144,
        "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
     }
     elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
        &logError ( 20145,
        "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
     }
     else {
       #----------------
       # Re-initialise..
       #----------------
       my $sGrpName = $lsHaresOut[0] ;
       &logTrace ( 20146,
          "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
       my $sHagrpCmdArgs = "-value $sGrpName Frozen";
  
       #---------------------------------------------
       # Assume the group name is valid and proceed..
       #---------------------------------------------
       my @lsHagrpOut = grep { s/\n// || 1 } readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
  
       #-----------------------------------------
       # Check if command executed successfully..
       #-----------------------------------------
       if ( $iExitCode = $? >> 8 ) {
          &logError ( 20147,
          "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
       }
       elsif ( $lsHaresOut[0] =~ /10048.*Attribute Frozen does not exist/o ) {
          &logError ( 20148,
          "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
       }
       elsif ($lsHagrpOut[0] ne "1" ){
         &logWarn ( 20141,
         "<%s> Missing sapstartsrv process in first level monitor.Marking for Restart.",
         $sMyName );
         if ( SAPNW::StartSapService ( $sInstProfile,
                                       $sSAPAdmin,
                                       $sEnvFile,
                                       $bUseSystemD ) ) {
           &logInfo ( 20142,
           "<%s> sapstartsrv started successfully.", $sMyName );
         }
         else {
           &logWarn ( 20143,
           "<%s> sapstartsrv could not be started.", $sMyName );
         }
       }
    }
  }

  if ( ! $bErFlag                                             &&          # Missing er..
       $bSapstartSrvFlag                                      &&          #sapstartsrv is present
       ( $sInstType =~ /^(enqrep)$/io )                       && 
       $sEntryPointName =~ /^(monitor|imf_register)$/o        &&          # In monitor..
       $sProcMon !~ /^(JSTART|JSTARTM|JC|jlaunch)$/o          &&          # Not in recursive call..
       ( VCSDef_ResourceState::ONLINE        == $::giState )  &&          # Resource is ONLINE
       ( VCSDef_ResourceIState::NOT_WAITING  == $::giIState ) &&          # Resource is in steady ONLINE state
       VCSDef_MonitorExit::OFFLINE == $iReturnCode ) { # offline..
      my $sHaresCmd = $sCmdHares;
      my $sHagrpCmd = $sCmdHagrp;
      my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
      my $iExitCode = () ;
     #----------------
     # Re-initialise..
     #----------------
     my $sHaresCmdArgs = "-value $sResourceName Group";
  
     #---------------------------------------------
     # Assume the group name is valid and proceed..
     #---------------------------------------------
     my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
     #-----------------------------------------
     # Check if command executed successfully..
     #-----------------------------------------
     if ( $iExitCode = $? >> 8 ) {
        &logError ( 20144,
        "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
     }
     elsif ( $lsHaresOut[0] =~ /10048.*Attribute Group does not exist/o ) {
        &logError ( 20145,
        "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
        $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
     }
     else {
       #----------------
       # Re-initialise..
       #----------------
       my $sGrpName = $lsHaresOut[0] ;
       &logTrace ( 20146,
          "<%s> Group for resource %s is %s.", $sMyName,$sResourceName,$sGrpName );
       my $sHagrpCmdArgs = "-value $sGrpName Frozen";
  
       #---------------------------------------------
       # Assume the group name is valid and proceed..
       #---------------------------------------------
       my @lsHagrpOut = grep { s/\n// || 1 } readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
  
       #-----------------------------------------
       # Check if command executed successfully..
       #-----------------------------------------
       if ( $iExitCode = $? >> 8 ) {
          &logError ( 20147,
          "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
       }
       elsif ( $lsHaresOut[0] =~ /10048.*Attribute Frozen does not exist/o ) {
          &logError ( 20148,
          "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
          $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
       }
       elsif ($lsHagrpOut[0] ne "1" ){
       &logTrace ( 20149,
          "<%s> Group for resource %s is %s and is not Frozen.", $sMyName,$sResourceName,$sGrpName );
       if( ( VCSDef_ResourceState::ONLINE   == $::giState ) && 
           ( VCSDef_ResourceIState::WAITING_TO_GO_OFFLINE  != $::giIState ))
       {
         #----------------
         # Re-initialise..
         #----------------
         $sHaresCmdArgs = "-value $sResourceName EnqSrvResName";
  
         #---------------------------------------------
         # Assume the group name is valid and proceed..
         #---------------------------------------------
         my @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sHaresCmd $sHaresCmdArgs" );
  
         #-----------------------------------------
         # Check if command executed successfully..
         #-----------------------------------------
         if ( $iExitCode = $? >> 8 ) {
            &logError ( 20150,
            "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
            $sMyName, $sHaresCmd, $sHaresCmdArgs, $iExitCode );
         }
         elsif ( $lsHaresOut[0] =~ /10048.*Attribute EnqSrvResName does not exist/o ) {
            &logError ( 20151,
            "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
            $sMyName, $sHaresCmd, $sHaresCmdArgs,  $lsHaresOut[0] );
         }
         else {
           my $sEnqSrvResName = $lsHaresOut[0];
           if( $sEnqSrvResName ne "" ) {
             $sHaresCmd = $sCmdHares;
             my $sHasyscmd = $sCmdHasys;
             my $sHasyscmdArgs = " -nodeid ";
             my $sHasysOut = readpipe ( "$sHasyscmd $sHasyscmdArgs" );
             chomp $sHasysOut;
             $sHasysOut =~ s/.*Node name: //;
             $sHasysOut =~ s/ Node ID: .*//;


             # call hares display and look to get the IState info

             #call display with resource type
             my $sAttributes = "IState State";
             my $sGroups = "";
             my $sTypes = "SAPNW";
             my $sSystems = $sHasysOut;
             my $bIsEnqueueResWaitingOnline = 0;
             my $bIsEnqueueResOnline = 0;

             my @lsResultsList = ();
             if(Hares::Display($sEnqSrvResName,
                                 $sAttributes,
                                 $sGroups,
                                 $sTypes,
                                 $sSystems,
                                 \@lsResultsList)) {

               #ran display command ok - continue
               #spin the list and look at IState and State info
               foreach my $sEachResult (@lsResultsList) {

                 my @lsEachResultList = split(/\s/, $sEachResult);

                 my $sTmpResourceName = shift(@lsEachResultList);
                 my $sTmpAttribute = shift(@lsEachResultList);
                 my $sTmpSystem = shift(@lsEachResultList);
                 my $sTmpValue = join(" ", @lsEachResultList);


                 my $sTmpValueLC = lc($sTmpValue);

                 if(($sTmpAttribute  eq "IState") && ($sTmpValueLC =~ "waiting") && ($sTmpValueLC =~ "online")) {
                   #----------------
                   # Re-initialise..
                   #----------------
                   $bIsEnqueueResWaitingOnline = 1 ;
                   $sHaresCmdArgs = " -wait $sEnqSrvResName State ONLINE -sys $sHasysOut -time 300";

                   #---------------------------------------------
                   # Assume the group name is valid and proceed..
                   #---------------------------------------------
                   $sHaresCmd = $sHaresCmd . $sHaresCmdArgs ;
                   &logInfo ( 20152,
                     "<%s> Executing command %s.", $sMyName ,$sHaresCmd);
                   $iExitCode = system($sHaresCmd) ;

                   #-----------------------------------------
                   # Check if command is not executed successfully..
                   #-----------------------------------------
                   if ( $iExitCode == 0 ) {
                     $bIsEnqueueResOnline = 1;
                     $bIsEnqueueResWaitingOnline = 0 ;
                   }
                 }
                 if(($sTmpAttribute  eq "State") && ($sTmpValueLC =~ "online") ) {
                   $bIsEnqueueResOnline = 1;
                 }
               }
  
  
             #-----------------------------------------
             # Check if command is not executed successfully..
             #-----------------------------------------
             if ( ($bIsEnqueueResOnline != 1 ) && ($bIsEnqueueResWaitingOnline != 1 ) ) {
                &logWarn ( 20153,
                "<%s> Missing er process in first level monitor.Marking for Restart.",
                $sMyName );
                if ( SAPNW::StartERSInstance ( $sInstProfile,
                                              $sSAPAdmin,
                                              $sEnvFile,
                                              $bUseSystemD) ) {
                  &logInfo ( 20154,
                  "<%s> er started successfully.", $sMyName );
                  &logWarn ( 20155,
                  "<%s> Deleting PID files for VCS resource [%s]",
                  $sMyName, $sResourceName );
                  SAPNW::DeletePidFile ( $sResourceName );
                  #--------------------------------------
                  # Call ourselves again with 'er' ..
                  #--------------------------------------
                  $iFound = SAPNW::GetProcState ( \@liDummyPids,
                                                $sInstProfile,
                                                $sInstType,
                                                'er',
                                                $sSAPAdmin,
                                                $sEnvFile,
                                                0,               # Regular FLM..
                                                $bStopService,
                                                $bUseSystemD );
                   $iReturnCode = $iFound;
                   @$rliOnlinePids = @liDummyPids;
                }
                else {
                  &logWarn ( 20156,
                  "<%s> er could not be started.", $sMyName );
                }
             }
            }
           }
         } 
        }
       }
      }
    }


  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20157,
    "<%s> SAPNW process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @$rliOnlinePids;

  &logTrace ( 20158,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String ProcMon,
                          IN String InstProfile,
                          IN String InstType,
                          IN String MonitorProgram,
                          IN String SAPAdmin,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of Processes to be monitored 
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sMonitorProgram           Full path to external monitor program
 $sSAPAdmin                 SAP System Administrator
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sInstType,
       $sMonitorProgram,
       $sSAPAdmin,
       $iSecondLevelMonitor,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20159, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  MonitorProgram           [%s]
  SAPAdmin                 [%s]
  SecondLevelMonitor       [%s]
  level_two                [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, $sInstType, 
  $sMonitorProgram, $sSAPAdmin, $iSecondLevelMonitor, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( -300 == $level_two ) {
    $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );
    &logTrace ( 20160, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }

  if ($level_two >= 1  ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPNW::DelayedAttributeCheck ( $sEnvFile,
                                             $sInstProfile,
                                             $sSAPAdmin ) ) ) {
      &logTrace ( 20161,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20162,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20163,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPNW::SecondLevelMonitor ( $sEnvFile,
                                          $sProcMon,
                                          $sInstProfile,
                                          $sInstType,
                                          $sSAPAdmin,
                                          $iTillWhen ) ) ) {

      &logTrace ( 20164,
      "<%s> SAPNW instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20165,
      "<%s> SAPNW instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + SAPNW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPNW::ExternalMonitor ( $sSAPAdmin,
                                       $sEnvFile,
                                       $sExtCmd,
                                       $sExtCmdArgs,
                                       $iTillWhen ) ) ) {
      &logTrace ( 20166,
      "<%s> SAPNW instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20167,
      "<%s> SAPNW instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String InstProfile,
                                        IN String SAPAdmin );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sInstProfile,
       $sSAPAdmin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20168, "Subroutine <%s> called with:
  EnvFile                  [%s]
  InstProfile              [%s]
  SAPAdmin                 [%s]",

  $sMyName, $sEnvFile, $sInstProfile, $sSAPAdmin );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile InstProfile SAPAdmin );
  my @lsReqdArgs = qw ( EnvFile InstProfile SAPAdmin );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20169,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #-------------------------------
  # Validate parameter SAPAdmin..
  #-------------------------------
  if ( ! defined getpwnam $sSAPAdmin ) {
    &logError ( 20170, "<%s> SAPAdmin [%s] is not a valid UNIX login",
    $sMyName, $sSAPAdmin );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20171, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------------------
  # Validate parameter InstProfile..
  #----------------------------------
  if ( ! -e $sInstProfile ) {
    &logError ( 20172, "<%s> InstProfile [%s] file does not exist",
    $sMyName, $sInstProfile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String ProcMon,
                                     IN String InstProfile,
                                     IN String InstType,
                                     IN String SAPAdmin,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of processes to be monitored
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sInstType,
       $sSAPAdmin,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20173, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, $sInstType,
  $sSAPAdmin, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  my ($sSAPSID, $sInstName, $iInstID, $sSAPHost, $sSAPHome) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my ( $sSwitch, $iRetCode ) = ();

  if ( $sInstType =~ /^(sapstartsrv)$/io ) {
    &logTrace ( 20174,
    "<%s> Skipping level two monitor for InstType [%s] InstanceName [%s] and SID [%s]",
    $sMyName, $sInstType, $sInstName, $sSAPSID );
    return VCSDef_MonitorExit::ONLINE;
  }

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if( 1 == grep { /^(jc)$/o } @lsProcMon ) {
    $sArch = 1;
  }
  elsif ( 2 == grep { /^(jstart|dw)$/o } @lsProcMon ) {
    $sArch = 3;
  }
  elsif ( 1 == grep { /^(dw)$/o } @lsProcMon ) {
    $sArch = 2;
  }

  if ( $hsArch { $sArch } ) {
    $sSwitch = $hsArch { $sArch };
    &logTrace ( 20174,
    "<%s> Identified SAP Architecture [%s] from ProcMon [%s]",
    $sMyName, $sSwitch, $sProcMon );
  }
  else {
    &logTrace ( 20175,
    "<%s> Could not identify SAP Architecture from ProcMon [%s]",
    $sMyName, $sProcMon );
    $sSwitch = 'DEFAULT'; # Mark it default..
  }

  #-----------------------
  # Do the correct stuff..
  #-----------------------

  if("Java" eq $sSwitch ) { 

      if ( $sInstType =~ /^(appserv|smdagent)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function J2EEGetProcessList";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPNW::ExecGetProcList ( $sSAPAdmin,
                                               $sSwitch,
                                               $sProcMon,
                                               $sEnvFile,
                                               $sCommand,
                                               $sCommandArgs,
                                               $iTimeout );
      }
      else {
        &logError ( 20176,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

  }

  elsif("ABAP" eq $sSwitch ) { 

      if ( $sInstType =~ /^(appserv)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function GetProcessList";
        $iTimeout = $iTillWhen - time;

        $iReturnCode = SAPNW::ExecGetProcList ( $sSAPAdmin, # User context..
                                                 $sSwitch,
                                                 $sProcMon,
                                                 $sEnvFile,  # Use envfile..
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iTimeout );
      }
      else {
        &logError ( 20177,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

  }

  elsif("AddIn" eq $sSwitch ) { 

      if ( $sInstType =~ /^(appserv)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function GetProcessList";
        $iTimeout = $iTillWhen - time;
	
        $iReturnCode = SAPNW::ExecGetProcList ( $sSAPAdmin, # User context..
                                                 'ABAP',
                                                 $sProcMon,
                                                 $sEnvFile,  # Use envfile..
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
      }
      if ( $sInstType =~ /^(appserv|smdagent)$/io ) {  
        #--------------------------
        # Re-set command and args..
        #--------------------------

        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function J2EEGetProcessList";
        $iTimeout = $iTillWhen - time;

        $iRetCode = SAPNW::ExecGetProcList ( $sSAPAdmin,
                                            'Java',
                                            $sProcMon,
                                            $sEnvFile,
                                            $sCommand,
                                            $sCommandArgs,
                                            $iTimeout );
        #-------------------------------------------
        # Update, status only when it's not online..
        #-------------------------------------------
        unless ( VCSDef_MonitorExit::ONLINE == $iRetCode ) {
          $iReturnCode = $iRetCode;
        }
      }
      else {
        &logError ( 20178,
        "<%s> Encountered unsupported InstType [%s] for SAP Arch [%s]",
        $sMyName, $sInstType, $sSwitch );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

  }

  else{ 

      if ( $sInstType =~ /^(hana)$/io ) {
        #-----------------------
        # Set command and args..
        #-----------------------
        $sCommand = "$sSAPHome/sapcontrol";
        $sCommandArgs = " -nr $iInstID -function GetProcessList";
        $iTimeout = $iTillWhen - time;
	
        $iReturnCode = SAPNW::ExecGetProcList ( $sSAPAdmin, # User context..
                                                 'HANA',
                                                 $sProcMon,
                                                 $sEnvFile,  # Use envfile..
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iTimeout );

        #---------------------
        # If offline, return..
        #---------------------
        if ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ||
             VCSDef_MonitorExit::TIMEOUT == $iReturnCode ) {
          return $iReturnCode;
        }
      }
      elsif ( $sInstType =~ /^(enqueue|enqrep)$/io ) {
        
        if ( $hsProcs { enq } || $hsProcs { enqr } ) {
          $sCommand = "$sSAPHome/enq_admin";
          $sCommandArgs = " --replication_state pf=$sInstProfile ";
          $iTimeout = $iTillWhen - time;
 
          my $iEnqAdminReturnCode = SAPNW::ExecEnqAdmin ( $sEnvFile,
                                                          $sSAPAdmin,
                                                          $sCommand,
                                                          $sCommandArgs,
                                                          $iTimeout );
          if ( $iEnqAdminReturnCode == 3 ) {
            &logInfo ( 20471, "<%s> Enqueue Server 2 can not be reached. enq_admin command returned [%s]", $sMyName, $iEnqAdminReturnCode );
          } else {
            &logTrace ( 20472, "<%s> enq_admin command returned [%s]", $sMyName, $iEnqAdminReturnCode );
          }
        }
        else {
          #-----------------------
          # Set command and args..
          #-----------------------
          $sCommand = "$sSAPHome/ensmon";
          $sCommandArgs = "pf=$sInstProfile 2";
          $iTimeout = $iTillWhen - time;
         
          $iReturnCode = SAPNW::ExecENSMon ( $sSAPAdmin,
                                                  $sEnvFile,
                                                  $sCommand,
                                                  $sCommandArgs,
                                                  $sInstType,
                                                  $iTimeout );
        }
      }
      else {
        &logError ( 20179,
        "<%s> Encountered unsupported InstType [%s]. Has to be one of
        [enqueue enqrep]",
        $sMyName, $sInstType );
        $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
      }

  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String SAPAdmin,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20180, "Subroutine <%s> called with:
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20181,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20182,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20183,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20184,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20185,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPNW::ExecLocalCmd ( $sSAPAdmin, 
                             $sEnvFile, 
                             $sCmd, 
                             $sCmdArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20186,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20187,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20188, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20189, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20190,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20191,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20192,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool DeletePidFile ( IN String ResourceName );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub DeletePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20193, "Subroutine <%s> called with:
  ResourceName        [%s]",

  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20194,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "$VCSLOG/log/tmp/.VRTSSAPNW/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20195,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20196,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20197,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20198, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20199, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20200, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20201, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20202,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20203,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20204,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPNW::GetPidDetails ( $iPid,
                                     \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20205,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20206,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20207,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20208,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20209, "<%s> PID [%s] from PID file [%s] does not exist",
    $sMyName, $iPid, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPNWProcesses ( IN String InstProfile,
                                           IN String InstType,
                                           IN String ProcMon,
                                           IN String SAPAdmin,
                                           IN String EnvFile,
                                           IN Bool   StopService,
                                           IN Bool   UseSystemD );

 Description  : Kills processes for this instance of SAPNW
 
 Parameters :

 $sInstProfile            Instance profile file for the SAP Instance
 $sInstType               Instance type for the SAP Instance
 $sProcMon                List of processes to be monitored
 $sSAPAdmin               SAP System Administrator
 $sEnvFile                SAP Environment File 
 $bStopService            Specifies whether SAP Start Service should be stopped or not.
 $bUseSystemD             When enabled, resource comes Online in system.slice
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillSAPNWProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sInstType,
       $sProcMon,
       $sSAPAdmin,
       $sEnvFile,
       $bStopService,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20210, "Subroutine <%s> called with:
  InstProfile              [%s]
  InstType                 [%s]
  ProcMon                  [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  StopService              [%s]
  UseSystemD               [%s]",

  $sMyName, $sInstProfile, $sInstType, $sProcMon, $sSAPAdmin, $sEnvFile, 
  $bStopService, $bUseSystemD );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPNW::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPNW::GetProcState ( \@liGoners,
                               $sInstProfile,
                               $sInstType,
                               $sProcMon,
                               $sSAPAdmin,
                               $sEnvFile,
                               0,             # Set filtering choice..
                               $bStopService,
                               $bUseSystemD );
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20211,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20212,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20213,
    "<%s> Could not locate any process this instance of SAPNW",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20214, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20215,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20216,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20217,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20218,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20219,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20220,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20221,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20222,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20223, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20224,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if("solaris" eq $sSwitch) {

      my $sOSVer = ( uname ) [2]; # Get OS version..
      if ($sOSVer =~ /\d+\.(\d+)/) {
	$sOSVer = $1;
      }
      $sPsCmd = "/usr/bin/ps";
      if ( $sOSVer < 11 ) {
        $sPsCmd = "/usr/ucb/ps" if ( -x "/usr/ucb/ps" );
      }
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }

  elsif("hpux" eq $sSwitch) {

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  elsif("aix" eq $sSwitch) {

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }


  elsif("linux" eq $sSwitch) {

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  &logTrace ( 20225, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20226,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20227,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    &logWarn ( 20228,
    "<%s> Deleting PID files for VCS resource [%s]",
    $sMyName, $sResourceName );

    SAPNW::DeletePidFile ( $sResourceName );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String InstProfile,
                                  IN String ProcMon,
                                  IN Bool NWVersionLt71,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sInstProfile              Instance profile file for the SAP Instance
 $sProcMon                  List of processes to be monitored
 $bNWVersionLt71            boolean value which state whether SAP NetWeaver version is less than 7.1
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sProcMon,
       $bNWVersionLt71,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20229, "Subroutine <%s> called with:
  InstProfile              [%s]
  ProcMon                  [%s]
  NWVersionLt71            [%s]
  SrvrDetails              [%s]",

  $sMyName, $sInstProfile, $sProcMon, $bNWVersionLt71,$rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $sTmpProc, $sSAPSID, $sInstName,
       $iInstID, $sSAPHost, $sSAPHome, @lsWords ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "$VCSLOG/log/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20230, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20231, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  #----------------------
  # Decide architecture..
  #----------------------
  ( @lsWords ) = split( /\//, $sInstProfile );
  ( $sSAPSID, $sInstName, $sSAPHost ) = split ( /_/, $lsWords [-1] );
  ( $iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  my $sArch = 0;
  my %hsArch = ( 1 => 'Java',
                 2 => 'ABAP',
                 3 => 'AddIn' );

  my @lsProcMon = $sProcMon =~ /([^\s]+)/g;
  my %hsProcs = map { $_ => 1 } @lsProcMon;

  #--------------------------------
  # Identify the SAP architecture..
  #--------------------------------
  if ( $bNWVersionLt71 && $hsProcs { jc } && $hsProcs { dw } ) {
    $sArch = 3;
  }
  elsif ( $hsProcs { jc } ) {
    $sArch = 1;
  }
  elsif ( $hsProcs { jstart } && $hsProcs { dw } ) {
    $sArch = 3;
  }
  elsif ( $hsProcs { dw } ) {
    $sArch = 2;
  }

=pod
  if( 1 == grep { /^(jc)$/o } @lsProcMon ) {
    $sArch = 1;
  }
  elsif ( 2 == grep { /^(jstart|dw)$/o } @lsProcMon ) {
    $sArch = 3;
  }
  elsif ( 1 == grep { /^(dw)$/o } @lsProcMon ) {
    $sArch = 2;
  }
=cut

  if ( $sArch ) {
    &logTrace ( 20232,
    "<%s> SAP Instance Architecture is [%s]",
    $sMyName, $hsArch { $sArch } );
  }
  else {
    &logTrace ( 20233,
    "<%s> SAP Instance is [%s]",
    $sMyName, $sInstName );
  }
  if( $bNWVersionLt71) {
     push @lsProcMon, 'jlaunch' if ( $sArch & 1 ); # SAP Architecture is Java,jc located..
  }
  else {
    if ( 1 == $sArch ) { # jc located..
      push @lsProcMon, 'jstart';
    }
    elsif ( 3 == $sArch ) {  # jstart located..
      push @lsProcMon, 'jstartm';
    }
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsProcMon ) {
    $sKey = "srchstr_${sProc}";
    $sPidFile = "$sPIDDir/pidfile.$sKey";

    if ( $sProc eq 'sapstart' ) {
      $sTmpProc = qr|$sProc\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif ( $sProc eq 'sapstartsrv' ) {
      $sTmpProc = qr|$sProc\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif ( $bNWVersionLt71 && ( $sProc eq 'jc' ) && ( 3 == $sArch ) ) { 
      $sTmpProc = qr|jcontrol\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif ( $bNWVersionLt71 && ( $sProc eq 'jlaunch' ) ) { 
	    $sTmpProc = qr|$sProc\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}.*server|;
    }
    elsif ( $sProc eq 'jstart' ) { # Child.. Atre 27 Mar 2009
      $sTmpProc = qr|$sProc\s.*\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}.*-hostvm|;
    }
    elsif ( $sProc eq 'jstartm' ) { # Parent.. Atre 27 Mar 2009
      $sTmpProc = qr|jstart\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif( $sProc eq 'icman' ){
      $sTmpProc = qr|$sProc\s.*\spf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif( $sProc eq 'gwrd' ){
      $sTmpProc = qr|$sProc(\s\|.sap${sSAPSID}_${sInstName}).*pf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }
    elsif( $sProc eq 'hdbnameserver' || $sProc eq 'hdbindexserver'  || $sProc eq 'hdbpreprocessor' || $sProc eq 'hdbcompileserver' || $sProc eq 'hdbstatisticsserver' || $sProc eq 'hdbxsengine') {
      $sTmpProc = qr|$sProc|;
    }
    elsif ( $sProc eq 'sapwebdisp_hdb' ) {
      $sTmpProc = qr|$sProc\spf=.*wdisp/sapwebdisp.pfl\s-f\s.*/dev_webdisp|;
    }
    else {
      $sTmpProc = qr|$sProc.sap${sSAPSID}_${sInstName}.*pf=.*profile/(START\|${sSAPSID})_${sInstName}_${sSAPHost}|;
    }

    @{ $$rhlSrvrDetails { $sProc } } = ( $sPidFile, $sTmpProc ); # Store..

    &logTrace ( 20234, "<%s> Generating filter [%s] for process [%s] with PID file [%s]",
    $sMyName, $sTmpProc, $sProc, $sPidFile );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20235, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 ,
		        'Imf_registerTimeout',10 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  if ( $sEntryPointName eq "imf_register" ) {
   $sEntryPointName = "monitor" ;
  }

  if ( lc $sEntryPointName ne "preonline" ) {

    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sCommandArgs = "-value $sResourceName $sAttribute";

    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20236,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20237, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20238, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  if ( lc $sEntryPointName eq "preonline" ) {
   $sEntryPointName = "online" ;
  }
  unless ( $iEngineTimeout ) {
    $sCommand = $sCmdHatype; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sResourceType = "SAPNW" if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20239,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    elsif ( ! $iEngineTimeout ) { # Empty value..
      &logTrace ( 20240, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    else { # Success..
      &logTrace ( 20241, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20242,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : bool UpdateSAPNWLog ( IN String InstName,
                                       IN String InstProfile,
                                       IN String InstType,
                                       IN String SAPAdmin,
                                       IN String SAPSID,
                                       IN String MessageText );

 Description : Update SAP log with passed message string..

 Parameters :

 $sInstName         String identifier that uniquely identifies a SAP server instance
 $sInstProfile      Specifies the full path to the Instance profile for this instance
 $sInstType         String identifier that classifies the SAP server instance type
 $sSAPAdmin         UNIX user dedicated for this instance of SAP
 $sSAPSID           SAP system name defined during installation
 $sMessageText      Message text for update


 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub UpdateSAPNWLog {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstName,
       $sInstProfile,
       $sInstType,
       $sSAPAdmin,
       $sSAPSID,
       $sMessageText ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20243, "Subroutine <%s> called with:
  InstName                 [%s]
  InstProfile              [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  SAPSID                   [%s]
  MessageText              [%s]",

  $sMyName, $sInstName, $sInstProfile,  $sInstType,
  $sSAPAdmin, $sSAPSID, $sMessageText );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sDate = strftime '%H:%M:%S %d %b %Y', localtime;
  my $sHost = ( uname ) [1];
  my $sLogFile = "/usr/sap/$sSAPSID/$sInstName/log/" .
                 "${sSAPSID}_${sInstName}.log";

  my $sCpyHeader = '+' . '-' x 75 . '+';
  my $sCmnt = "|";

  #------------------------
  # Define logging format..
  #------------------------

  format SAPLOG =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
@ Date: @<<<<<<<<<<<<<<<<<<<<< Node: @<<<<<<<<<<<<<<<<<<                    @
$sCmnt, $sDate, $sHost, $sCmnt
@ InstName: @<<<<<<<<<<<< InstType: @<<<<<<<<<< SAPSID: @<<<                @
$sCmnt, $sInstName, $sInstType, $sSAPSID, $sCmnt
@ Inst profile: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @
$sCmnt, $sInstProfile, $sCmnt
@ @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @
$sCmnt, $sMessageText, $sCmnt
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sCpyHeader
.

  #-----------------------
  # Open file for update..
  #-----------------------
  if ( sysopen ( SAPLOG, $sLogFile, O_CREAT|O_WRONLY|O_APPEND|O_SYNC, 0644 ) ) {
    &logTrace ( 20244,
    "<%s> Opened log file [%s] successfully for update", $sMyName, $sLogFile );

    write SAPLOG; # Update..

    unless ( close ( SAPLOG ) ) {
      &logError ( 20245,
      "<%s> Failed to close log file [%s] after update: [%s]",
      $sMyName, $sLogFile, $! );
      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20246,
    "<%s> Failed to open log file [%s] for update: [%s]",
    $sMyName, $sLogFile, $! );
    $iReturnCode = 0; # Failure..
  }

  #----------------
  # Set ownership..
  #----------------
  my ( $iUid, $iGid ) = ( getpwnam $sSAPAdmin ) [2, 3]; # Should not fail..

  if ( chown ( $iUid, $iGid, $sLogFile ) ) {
    &logTrace ( 20247,
    "<%s> Set ownership of [%s] to UID [%s] GID [%s]",
    $sMyName, $sLogFile, $iUid, $iGid );
  }
  else {
    &logError ( 20248,
    "<%s> Failed to set ownership of [%s] to UID [%s] GID [%s]: [%s]",
    $sMyName, $sLogFile, $iUid, $iGid, $! );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecGetProcList ( IN String SAPAdmin,
                            IN String SAPArch,
                            IN String ProcMon,
                            IN String EnvFile,
                            IN String Cmd,
                            IN String CmdArgs,
                            IN Integer Timeout );

 Description : Decides on status of SAP Instance based on sapcontrol command.

 Parameters :

 $sSAPAdmin           UNIX user dedicated for this instance of SAP
 $sSAPArch	          SAP Architecture which could be 'Java' or 'ABAP' or 'HANA'
 $sProcMon 	          List of processes to be monitored 
 $sEnvFile            Full path to the environment file for SAP
 $sCmd                Full path to sapcontrol binary
 $sCmdArgs            Argument list for sapcontrol
 $iTimeout            Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub ExecGetProcList {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sSAPArch,
       $sProcMon,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20249, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  SAPArch       [%s]
  ProcMon       [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sSAPArch, $sProcMon, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20250,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20251,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20252,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  $iReturnCode = SAPNW::ExecLocalCmd ( $sSAPAdmin, # User context..
                                            $sEnvFile,  # Use envfile..
                                            $sCmd,
                                            $sCmdArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdOut ); 

  if( defined $iExitCode ){
    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20253,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );

    #------------------------------------------------------
    # Get status based on state of dispatcher or servers or hana processes..
    #------------------------------------------------------
      $iReturnCode = SAPNW::GetSapControlStatus ( $sSAPArch, $sProcMon, \@lsCmdOut );
      @lsCmdOut = (); # Free..
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20254,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif( $iExitCode == 4 ) {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    &logError ( 20255,
    "<%s> Cmd[%s] executed successfully with exit code 4 indicating  all processes stopped",
    $sMyName, $sCmd );
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20256,
    "<%s> Cmd [%s] failed with exit code [%s]. Cmd output [%s]
    Returning [%s]", $sMyName, $sCmd, $iExitCode,
    join ( chr 012, @lsCmdOut ), $iReturnCode );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExecENSMon ( IN String SAPAdmin,
                             IN String EnvFile,
                             IN String Cmd,
                             IN String CmdArgs,
                             IN String InstType,
                             IN Integer Timeout );

 Description : Decides on status of SAP Instance based on ensmon command.

 Parameters :

 $sSAPAdmin   UNIX user dedicated for this instance of SAP
 $sEnvFile    Full path to the environment file for SAP
 $sCmd        Full path to sapinfo binary
 $sCmdArgs    Argument list for ensmon
 $sInstType   String identifier that classifies the SAP server instance type
 $iTimeout    Time in seconds for execution

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExecENSMon {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPAdmin,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $sInstType,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20257, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  InstType      [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $sInstType,
  $iTimeout );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sEntryPointName, $iExitCode, @lsCmdOut ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20258,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20259,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20260,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( SAPNW::ExecLocalCmd ( $sSAPAdmin, # User context..
                                     $sEnvFile,  # Use envfile..
                                     $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20261,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  elsif ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdOut [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..

    &logError ( 20262,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 4 == $iExitCode && $sInstType =~ /^enqueue$/io ) {
    if ( $sEntryPointName eq 'monitor' ) {
      &logCritical ( 20263,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    else {
      &logTrace ( 20264,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Replication Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  elsif ( 8 == $iExitCode ) {  # Up already..
    if ( $sInstType =~ /^enqueue$/io ) {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;

      &logError ( 20265,
      "<%s> Cmd [%s] with args [%s] exited with [%s]. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $iReturnCode );
    }
    elsif ( $sInstType =~ /^enqrep$/io ) {
      &logCritical ( 20266,
      "<%s> Cmd [%s] with args [%s] exited with [%s].
      Not able to connect to ENQUEUE Server!!",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
  }
  else { # Error..
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logError ( 20267,
    "<%s> Cmd [%s] failed with exit code [%s]. Returning [%s]",
    $sMyName, $sCmd, $iExitCode, $iReturnCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetSapControlStatus ( IN String SAPArch,
                                      IN String ProcMon,
                                      IN RefStrList CmdOut );

 Description : Decides on status of SAP Instance based on given sapcontrol
               output.

 Parameters :

 $sSAPArch            SAP Architecture which could be 'Java' or 'ABAP' or 'HANA'
 $sProcMon 	          List of processes to be monitored 
 $rlsCmdOut           Reference to list containing output from sapcontrol execution.

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetSapControlStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSAPArch, $sProcMon, $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20268, "Subroutine <%s> called with:
  sSAPArch         [%s],
  sProcMon         [%s],
  rlsCmdOut         [%s]",
  $sMyName, $sSAPArch, $sProcMon, $rlsCmdOut );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------------
  # Validate against empty list..
  #------------------------------
  return VCSDef_MonitorExit::UNKNOWN unless scalar @$rlsCmdOut;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sTmp, $sKey ) = ( '' ) x 2;
  my ( $iSrv, $iRunning ) = ( 0 ) x 2;
  if( "ABAP" eq $sSAPArch ) {
    foreach (  @$rlsCmdOut ) {
        if ( $_ =~ /\b(Dispatcher)\b\,\s+(.*?)\s*\d/io &&
             $2 =~ /Green\,\s+running/io ) {
          ++ $iSrv;
          ++ $iRunning;
          &logTrace ( 20269, "<%s> Dispatcher is in running state",$sMyName );
          last;
        }
    }  

  }
  elsif( "Java" eq $sSAPArch ) {
    my %hsCmdOut = map { /\b(smdagent|server\d*)\b\,\s*\d*(.*)/io } @$rlsCmdOut;

    foreach $sKey ( sort keys %hsCmdOut ) {
      $sTmp .= "Server Name [$sKey] State [$hsCmdOut{$sKey}]\n";
      if ( $sKey =~ /(smdagent|server)/io ) {
        ++ $iSrv;
        ++ $iRunning if $hsCmdOut { $sKey } =~ /Running/io;
      }
    }
  }  
  elsif( "HANA" eq $sSAPArch ) {
    my @sHanaProcMon= $sProcMon =~ /([^\s]+)/g;
    @sHanaProcMon = grep { lc $_  !~ /^hdb$/io } @sHanaProcMon;
    my $sHanaProcesses = join ( '|', @sHanaProcMon );
    my %hsCmdOut = map { /\b(hdbdaemon|hdbnameserver|hdbindexserver|$sHanaProcesses)\b\,\s*\d*(.*)/io } @$rlsCmdOut;

    foreach $sKey ( sort keys %hsCmdOut ) {
      $sTmp .= "Server Name [$sKey] State [$hsCmdOut{$sKey}]\n";
      if ( $sKey =~ /(hdbdaemon|hdbnameserver|hdbindexserver|$sHanaProcesses)/io ) {
        ++ $iSrv;
        ++ $iRunning if $hsCmdOut { $sKey } =~ /Green\,\s+running/io;
      }
    }
  }  
  &logTrace ( 20270, "<%s> SAP sapcontrol status : ---\n%s", $sMyName, $sTmp );

  #-----------------
  # Check on state..
  #-----------------
  if( ( "HANA" eq $sSAPArch ) && ( $iSrv != $iRunning ) ) {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logWarn ( 20272,
    "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
    $sMyName, $sTmp, $iReturnCode );
  }
  elsif ( $iRunning ) {                            # One server running..
    &logTrace ( 20271,
    "<%s> Found [%s] of [%s] server(s) running. Returning [%s]",
    $sMyName, $iRunning, $iSrv, $iReturnCode );

  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    &logWarn ( 20272,
    "<%s> Status of SAP instance undecided:\n%sReturning [%s]",
    $sMyName, $sTmp, $iReturnCode );
  }

  ( $sTmp, $sKey ) = (); # Free..

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ChkResName ( IN String ResName,
                              IN String SAPSID,
                              IN String InstType );

 Description : Checks if resource has been configured

 Parameters :

 $sResName         Resource Name to check
 $sSAPSID          SAP system name defined during installation
 $sInstType        String identifier that classifies the SAP server instance type

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ChkResName {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResName,
       $sSAPSID,
       $sInstType ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20273, "Subroutine <%s> called with:
  ResName            [%s]
  SAPSID             [%s]
  InstType           [%s]",

  $sMyName, $sResName, $sSAPSID, $sInstType );
  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1;

  #-------------------------------
  # Validate string for resource..
  #-------------------------------
  return 0 unless ( $sResName =~ /\S/o ); # Must be a non-white string..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..
  my $sCmdArgs = "-display $sResName -attribute InstProfile InstType";

  #-----------------------
  # Define valid matches..
  #-----------------------
  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE' );

  my ( $iExitCode, @lsCmd ) = ();

  &logTrace ( 20274,
  "<%s> Executing Cmd [%s] with Args [%s]",
   $sMyName, $sCmd, $sCmdArgs );
  #--------------------------
  # Fire command and verify..
  #--------------------------
  ( @lsCmd ) = grep {
                 ( s/\n// || 1 )   &&            # Chomp, and make true..
                 /InstProfile|InstType/ &&       # Contains these..
                 s/.*\s+//                       # Get last field..
               } readpipe ( "$sCmd $sCmdArgs" ); # From output..


  my ($localSAPSID, $localInstName, $localSAPHost ) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $lsCmd [0] );
  ($localSAPSID, $localInstName, $localSAPHost) = split(/_/, $lsWords[-1]);

  #--------------------
  # Collect exit code..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20275,
    "<%s> Cmd [%s] with Args [%s] failed with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ( 2 == scalar @lsCmd )                  &&          # Got both values..
          ( $sSAPSID eq $localSAPSID )            &&          # SAPSID is a match..
          defined $hsInstMatch { uc $sInstType }  &&          # InsType is defined..
          $hsInstMatch { uc $sInstType } eq uc $lsCmd [-1] ) { # And a correct pair..
    &logTrace ( 20276,
    "<%s> Resource [%s] defined in cluster with InstType [%s] SAPSID [%s]",
     $sMyName, $sResName, $lsCmd [-1], $localSAPSID );
  }
  else {
    &logError ( 20277,
    "<%s> Resource [%s] not a matching ENQUEUE variant for this resource.
    The resource is with InstType [%s] SAPSID [%s]",
    $sMyName, $sResName, $lsCmd [-1], $localSAPSID );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineENQGrp ( IN String Sys,
                                IN String Grp,
                                IN String Reason,
                                IN String FaultedNode );

 Description : Pre-online trigger for SAPNW

 Parameters :

 $sSys                System where the group is asked to be made online
 $sGrp                Group for which preonline has been invoked
 $sReason             Contains the string MANUAL or FAULT
 $sFaultedNode        Cluster node where the fault occurred due to FAULT

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineENQGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSys,
       $sGrp,
       $sReason,
       $sFaultedNode ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20278, "Subroutine <%s> called with:
  Sys             [%s]
  Grp             [%s]
  Reason          [%s]
  FaultedNode     [%s]",

  $sMyName,
  $sSys, $sGrp, $sReason, $sFaultedNode );


  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHares; # Get global..

  my %hsInstMatch = ( 'ENQREP'  => 'ENQUEUE' );

  my $sCmdArgs = ();

  my $sOSVer = ();

  #---------------------
  # Validate arguments..
  #---------------------
  if ( $sReason eq 'MANUAL' ) {
    &logInfo ( 20279,
    "<%s> Invoked during a normal switch-over or online.", $sMyName );
  }
  elsif ( $sSys =~ /\S/ && ! inet_aton ( $sSys ) ) {
    &logError ( 20280,
    "<%s> Destination cluster node [%s] cannot be resolved",
    $sMyName, $sSys );
    $iReturnCode = 0; # Error..
  }
  elsif ( $sFaultedNode ne 'UNDEF' && $sFaultedNode =~ /\S/ && ! inet_aton ( $sFaultedNode ) ) {
    &logError ( 20281,
    "<%s> Faulted cluster node [%s] cannot be resolved",
    $sMyName, $sFaultedNode );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -e $sCmd ) {
    &logError ( 20282,
    "<%s> Command [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20283,
    "<%s> Command [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Error..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------------------------------
  # Bug in VCS for FAULT. Verify group is really faulted..
  #-------------------------------------------------------
  # hasys -list SysState=RUNNING

  my $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
  my $iSleepTime = ( $iTillWhen - time ) / 10;
  my $iTimeRemain = undef;
  my $iSleepCount = 0;

  while ( 1 ) { 

    my ( $i, $iExitCode, $sRes, $sDisp, $sGroup, $sProcMon, $sNodeName,  $sEnqSrvRes,
       $sCmdArgs,   $sTempNode,    $sEnqRes,   $sInstType,   $sFailHost,
       $sKey, $sSAPSID,  $sSID, @lsHaresOut, @lsCpy, @lsValidNodes, @lsEnqReps,
       @lsEnques,        %hsEnqueues,     %hsFltd,           %hsEnqReps,
       %hsEnqRepGrps, %hsEnqRepHosts, %hsSwEnqReps,    %hsOnlineEnqueues ,
       @lsHasysRunning, @lsHasysFault, @lsHasysOut, %hsHasysRunning, 
       %hsHasysFault,@lsWords, $sInstProfile, $sInstName, $sSAPHost, $iNumClusNodes) = ();

    if( $iSleepCount > 0 ) {
      unless ( ( $iTimeRemain = $iTillWhen - time ) > 10 ) {
        &logError ( 20385,
        "<%s> Insufficient time [%s] sec to retry. Exiting ..", $sMyName, $iTimeRemain );
        $iReturnCode = 0; # Error..
        last;
      }
      sleep $iSleepTime; # Take a break..
    }

    $iSleepCount++;
    $sCmd = $sCmdHasys;
    $sCmdArgs = "-state -localclus";

    @lsHasysOut = readpipe ( "$sCmd $sCmdArgs" );
    shift @lsHasysOut;

    @lsHasysRunning = grep {
                    ( s/\n//  || 1 ) &&
                    ( s/.*:// || 1 ) &&   # for DR ...
                    ( s/^(\S+).*(RUNNING)$/$1/ )
                  } @lsHasysOut;

    @lsHasysFault = grep {
                    ( s/\n//  || 1 ) &&
                    ( s/.*:// || 1 ) &&   # for DR ...
                    ( s/^(\S+).*(FAULTED|EXITED)$/$1/ )
                  } @lsHasysOut;

    %hsHasysRunning = map { $_ => 1 } @lsHasysRunning;

    &logTrace ( 20284,
    "<%s> List of systems where VCS running [%s]", 
    $sMyName, join ( ' ', @lsHasysRunning ) );

    %hsHasysFault   = map { $_ => 1 } @lsHasysFault;

    &logTrace ( 20285,
    "<%s> List of systems where VCS Faulted [%s]", 
    $sMyName, join ( ' ', @lsHasysFault ) );

    $sCmd = $sCmdHares;
    $sCmdArgs = "-list Group=$sGrp State=FAULTED -localclus";

    @lsHaresOut = grep {
                    ( s/\n// || 1 ) &&
                    s/\s+.*//
                  } readpipe ( "$sCmd $sCmdArgs" );

    #-------------------
    # Verify exit code..
    #-------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20286,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Group [%s] does not have any faulted resources.",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, $sGrp );

      #-----------------------------
      # Check for node FAULT event..
      #-----------------------------
      if ( $sFaultedNode eq 'UNDEF' && $sReason eq 'FAULT' ) {
        &logWarn ( 20287,
        "<%s> Missing faulted node name.
        Possible node FAULT event detected", $sMyName );
      }
      elsif ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' ) {
        &logWarn ( 20288,
        "<%s> Missing source node name.", $sMyName );
      }
      elsif ( ($sOSVer = $^O eq 'solaris' ? ( uname ) [2] : '') eq "5.10" ) {
        #---------------------------------------------------------------
        # Add zone specific support for Enqueue when zone itself faults.
        #---------------------------------------------------------------
        &logInfo ( 20289,
        "<%s> Checking if the Group [%s] is configured for Solaris non-global zones",
        $sMyName, $sGrp );

        my $sOut = ();
        my @lsOut = ();
        my $iCount = ();

        $sCmdArgs = "-display -attribute ContainerName -group $sGrp -type SAPNW -localclus";
        @lsHaresOut = grep { (s/\n// || 1) } readpipe ( "$sCmd $sCmdArgs" );

        shift @lsHaresOut;

        foreach $sOut (@lsHaresOut) {
          @lsOut = split( '\s+', $sOut);
          $iCount = scalar @lsOut;
          if( (4 == $iCount) && ($lsOut[2] =~ /$sSys/ || $lsOut[2] =~ /global/) ) {
            &logWarn ( 20290,
            "<%s> ContainerName attribute is configured.
            Possible zone [%s] FAULT event detected", $sMyName, $lsOut[3] );

            last;
          }
        }
      }
    }
    else {
      &logTrace ( 20291,
      "<%s> Group [%s] has [%s] resource(s) in FAULTED state",
      $sMyName, $sGrp, join ( ' ', @lsHaresOut ) );
    }

    #----------------
    # Re-initialise..
    #----------------
    $sCmdArgs = "-display -type SAPNW -attribute " .
                "State Group EnqSrvResName InstProfile InstType ProcMon -localclus";

    #---------------------------------------------
    # Assume the group name is valid and proceed..
    #---------------------------------------------
    @lsHaresOut = grep { s/\n// || 1 } readpipe ( "$sCmd $sCmdArgs" );

    #-----------------------------------------
    # Check if command executed successfully..
    #-----------------------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20292,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      return 0; # Error..
    }
    elsif ( $lsHaresOut[0] =~ /10554.*No resource exists with/o ) {
      &logError ( 20293,
      "<%s> Cmd [%s] with args [%s] failed with reason [%s]",
      $sMyName, $sCmd, $sCmdArgs,  $lsHaresOut[0] );
      return 0; # Error..
    }
  
    $sFailHost = $sSys;    # Set to default..
    my $sSysCopy = $sSys;    # Set to default..
    push @lsHaresOut, "#"; # Insert trailing de-marker..
    @lsCpy = @lsHaresOut;  # Backup..
  
    #---------------------------------------------------
    # Mark all ENQUEUE/ENQREP resources in given group..
    #---------------------------------------------------
    @lsEnques =
    grep {
       ( s/\n// || 1 )                   && # Chomp, and always make true..
       ( /Group.*\s+(\S+)/               && # Mark the group name..
       ( $sGroup = $1 ) && 0 )           || # Store, and always make false..
       ( /ProcMon\s+(\S+)\s+(\S+.*)/     && # Mark the ProcMon
       ( $sProcMon = $2 ) && 0 )         || # Store, and always make false..
       ( m#InstProfile.*\s+.*/(...)#     && # Mark the SID..
       ( $sSID = $1 ) && 0 )             || # Store, and always make false..
       ( s/^(\S+).*InstType.*\s             # Mark resource..
       (ENQ(UEUE|REP))/$1:$2/ix          && # If ENQUEUE or ENQREP variant..
       ++ $i )                           || # Set flag..
       ( $i                              && # If flag..
       s/.*/sprintf ( "$sSID:$sGroup:$sProcMon",      # Print group..
        -- $i )/e )                         # Reset flag..
    } @lsCpy;
  
    &logTrace ( 20294,
    "<%s> Searching ENQUEUE resource in group [%s]", $sMyName, $sGrp );
  
    &logTrace ( 20295,
    "<%s> Enques values are [%s]", $sMyName, join(' ', @lsEnques) );
  
    #-------------------
    # Start processing..
    #-------------------
    while ( 2 == ( ( $sRes, $sGroup ) = splice ( @lsEnques, 0, 2 ) ) ) {
      ( $sRes, $sInstType ) = $sRes =~ /([^:]+)/g;
      ( $sSAPSID, $sGroup, $sProcMon ) = $sGroup =~ /([^:]+)/g;
  
      #-------------------------------
      # Passed group has ENQ(UE|REP)..
      #-------------------------------
      if ( $sGrp eq $sGroup ) {
        if ( $sInstType =~ /ENQUEUE/io ) {
          &logInfo ( 20296,
          "<%s> +------- Group [%s] has ENQUEUE variant -------+",
          $sMyName, $sGrp );
  
          #--------
          # Store..
          #--------
          push @{ $hsEnqueues { $sRes } },
          $sSAPSID, $sInstType, $sGroup, $sProcMon; # Store..
        }
        elsif ( $sInstType =~ /ENQREP/io ) {
          &logInfo ( 20297,
          "<%s> +------- Group [%s] has ENQREP variant -------+",
          $sMyName, $sGrp );
  
          #--------
          # Store..
          #--------
          push @{ $hsSwEnqReps { $sRes } },
          $sSAPSID, $sInstType, $sGroup; # Store..
        }
      }
  
      &logTrace ( 20298,
      "<%s> Resource [%s] SAPSID [%s] InstType [%s] found in Group [%s]",
      $sMyName, $sRes, $sSAPSID, $sInstType, $sGroup );
    }
  
    #-----------------------------------------------
    # Find online resources and list their details..
    #-----------------------------------------------
    @lsCpy = @lsHaresOut; # Re-initialise..
    $i = 0;
  
    &logTrace ( 20299,
    "<%s> Looking for all ONLINE SAPNW resources", $sMyName );
  
    @lsEnqReps =
  
    grep {
      ( s/\n// || 1 )                        && # Chomp, and always make true..
      ( /^(\S+).*Group.*\s+(\S+)/            && # Mark resource, and group name..
      ( $sGroup = "$1:$2" ) && ( $sSys="" )  && 0 )             || # Store, and always make false..
      ( /.*\s(\S+)\s+ONLINE$/ && ( $sSys="$1" ) && 0 ) ||
      ( s/.*\s(\S+)\s+ONLINE$/$1/ && ++ $i ) || # If online, get node name, set flag..
      ( $i                                   && # If flag..
      /(EnqSrvResName|InstProfile|InstType)\s*(global|$sSys)/   && # and contains one of these..
      s/.*\s+// )                            || # get the last field..
      ( $i                                   && # If flag, use demarker..
      s/#/sprintf ( $sGroup, -- $i )/e )        # to print the group name.. reset flag..
    } @lsCpy;
  
    &logTrace ( 20300,
    "<%s> EnqReps [%s]", $sMyName, join(' ', @lsEnqReps) );
  
    while ( 5 == ( ( $sNodeName, $sEnqSrvRes, $sInstProfile, $sInstType, $sGroup ) = splice ( @lsEnqReps, 0, 5 ) ) ) {
  
      $sNodeName =~ s/.*://; # Remove cluster component, under GCO..
      ( $sRes, $sGroup ) = $sGroup =~ /([^:]+)/g;
  
      #----------------------------------
      # Store only ENQ(UE|REP) variants..
      #----------------------------------
      (@lsWords) = split( /\//, $sInstProfile);
      ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  
      #----------------------------------
      # Store only ENQ(UE|REP) variants..
      #----------------------------------
      if ( $sInstType =~ /ENQREP/io ) {
        &logTrace ( 20301, "<%s> Located ENQREP variant. Resource [%s] details:
        Group [%s] Node [%s] EnqSrvResName [%s] InstType [%s] SAPSID [%s]",
        $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );
  
        push @{ $hsEnqReps { $sRes } },
        $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
      }
      elsif ( $sInstType =~ /ENQUEUE/io ) {
        $sEnqSrvRes = ""; # Truncate..
        &logTrace ( 20302, "<%s> Located ENQUEUE variant. Resource [%s] details:
        Group [%s] Node [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s]",
        $sMyName, $sRes, $sGroup, $sNodeName, $sEnqSrvRes, $sInstType, $sSAPSID );
  
        push @{ $hsOnlineEnqueues { $sRes } },
        $sGroup, $sEnqSrvRes, $sInstType, $sSAPSID, $sNodeName;
      }
    }
  
    #--------------------------------------------
    # Validate if group has ENQ(UE|REP) variant..
    #--------------------------------------------
    @lsCpy = @lsHaresOut; # Re-initialise.
  
    if ( keys %hsEnqueues ) {
      #------------------------------------
      # Get all valid nodes for the ENQUEUE
      #------------------------------------
      $sRes =  join ( '|', keys %hsEnqueues );
      $sDisp =  join ( ' ', keys %hsEnqueues );
      &logInfo ( 20303,
      "<%s> Failover Group [%s] has ENQUEUE variant resource(s) [%s].",
      $sMyName, $sGrp, join ( ' ' , keys %hsEnqueues ) );

      %hsFltd = map { $_ => 1 }                 # All faulted nodes..
      grep {                                    # Grep'ed here..
        ( s/\n// || 1 )                    &&   # Chomp, and always make true..
        s/^($sRes).*State\s+(\S+).*FAULTED/$2/  # Pick all nodes where FAULTed..
      } @lsCpy;                                 # From cmd output..
  
      @lsCpy = @lsHaresOut;                     # Re-initialise.
      @lsValidNodes =
      grep {
        ( s/\n// || 1 )                       &&  # Chomp, and always make true..
        s/^($sRes)\s+.*State\s+(\S+).*LINE/$2/    # Pick the node not FAULTed..
      } @lsCpy;                                   # From cmd output..

      @lsCpy = @lsValidNodes; # Re-use..
      @lsValidNodes = ();     # Truncate..

      foreach $sKey ( @lsCpy ) {
          $sKey =~ s/.*://; # Remove cluster component, under GCO..
          push ( @lsValidNodes, $sKey ) unless $hsFltd { $sKey };
      }
  
      &logTrace ( 20461,
      "<%s> Nodes on which Service Group not in FALUTED state [%s]",
      $sMyName, join(' ', @lsValidNodes) );
 
      ( $sRes ) = keys %hsEnqueues; # Take one..
 
      #--------------------
      # Remove duplicates..
      #--------------------
      %hsFltd = (); # Truncate..
      @lsValidNodes = grep { ! $hsFltd { $_ } ++ } @lsValidNodes;
  
      &logTrace ( 20462,
      "<%s> One of ENQ resource(s) [%s] 
  
      Details for ENQREP resource [%s]:
  
      InstType [%s] SAPSID [%s]. Possible failover nodes [%s]",
      $sMyName, $sDisp, $sRes,
      ${ $hsEnqueues { $sRes } } [1], ${ $hsEnqueues { $sRes } } [0],
      join ( ' ', @lsValidNodes ) );
  
      #--------------------------------------------------------
      # Check if the node on which service group is configured 
      # is Running at this moment..
      #--------------------------------------------------------
      @lsValidNodes = grep { $hsHasysRunning { $_ } } @lsValidNodes;
  
      &logTrace ( 20463,
      "<%s> Valid nodes for ENQ failover targets [%s]",
      $sMyName, join ( ' ', @lsValidNodes ) );
    

    } # end of hsEnqueues processing **************
    elsif ( keys %hsSwEnqReps ) {
      #------------------------------------------
      # Get all valid nodes for the faulted REP..
      #------------------------------------------
      @lsCpy = @lsHaresOut;          # Re-initialise.
      $sRes =  join ( '|', keys %hsSwEnqReps );
      $sDisp =  join ( ' ', keys %hsSwEnqReps );
  
      &logInfo ( 20304,
      "<%s> Failover Group [%s] has ENQREP variant resource(s) [%s]",
      $sMyName, $sGrp, $sDisp );
  
      %hsFltd = map { $_ => 1 }                 # All faulted nodes..
      grep {                                    # Grep'ed here..
        ( s/\n// || 1 )                    &&   # Chomp, and always make true..
        s/^($sRes).*State\s+(\S+).*FAULTED/$2/  # Pick all nodes where FAULTed..
      } @lsCpy;                                 # From cmd output..
  
      @lsCpy = @lsHaresOut;                     # Re-initialise.
      @lsValidNodes =
      grep {
        ( s/\n// || 1 )                       &&  # Chomp, and always make true..
        s/^($sRes)\s+.*State\s+(\S+).*LINE/$2/   ||  # Pick the node not FAULTed..
        s/^($sRes)\s+.*EnqSrvResName.*\s+(\S+)/::$2/ # And EnqSrvResName..
      } @lsCpy;                                   # From cmd output..

      #-------------------------------------------------
      # Now extract EnqSrvResName and configured nodes..
      #-------------------------------------------------
      @lsCpy = @lsValidNodes; # Re-use..
      @lsValidNodes = ();     # Truncate..
  
      foreach $sKey ( @lsCpy ) {
        if ( $sKey =~ /^::/ ) { # EnqResName found..
          ( $sEnqRes = $sKey ) =~ s/:://;
        }
        else {
          $sKey =~ s/.*://; # Remove cluster component, under GCO..
          push ( @lsValidNodes, $sKey ) unless $hsFltd { $sKey };
        }
      }
  
      &logTrace ( 20305,
      "<%s> Nodes on which Service Group not in FALUTED state [%s]",
      $sMyName, join(' ', @lsValidNodes) );
  
      ( $sRes ) = keys %hsSwEnqReps; # Take one..
  
      #--------------------
      # Remove duplicates..
      #--------------------
      %hsFltd = (); # Truncate..
      @lsValidNodes = grep { ! $hsFltd { $_ } ++ } @lsValidNodes;
  
      &logTrace ( 20306,
      "<%s> One of ENQREP resource(s) [%s] defines EnqSrvResName as [%s].
  
      Details for ENQREP resource [%s]:
  
      InstType [%s] SAPSID [%s]. Possible failover nodes [%s]",
      $sMyName, $sDisp, $sEnqRes, $sRes,
      ${ $hsSwEnqReps { $sRes } } [1], ${ $hsSwEnqReps { $sRes } } [0],
      join ( ' ', @lsValidNodes ) );
  
      #--------------------------------------------------------
      # Check if the node on which service group is configured 
      # is Running at this moment..
      #--------------------------------------------------------
      @lsValidNodes = grep { $hsHasysRunning { $_ } } @lsValidNodes;
  
      &logInfo ( 20307,
      "<%s> Valid nodes for ENQREP failover targets [%s]",
      $sMyName, join ( ' ', @lsValidNodes ) );
    
      #--------------------------
      # If online ENQUEUE found..
      #--------------------------
      if ( defined $hsOnlineEnqueues { $sEnqRes } ) {
        &logTrace ( 20308,
        "<%s> Details for defined ENQUEUE resource [%s]:
        InstType [%s] SAPSID [%s] online on node [%s]",
        $sMyName, $sEnqRes, ${ $hsOnlineEnqueues { $sEnqRes } } [2],
        ${ $hsOnlineEnqueues { $sEnqRes } } [-2],
        ${ $hsOnlineEnqueues { $sEnqRes } } [-1] );
      }
      else {
        &logWarn ( 20389,
        "<%s> Corresponding ENQUEUE resource(s) is NOT online in the cluster. Retrying..", $sMyName );
        next;
      }

      #-----------------------------
      # Now validate EnqSrvResName..
      #-----------------------------
      if ( defined $hsOnlineEnqueues { $sEnqRes }             && # Defined..
           ( $hsInstMatch { (uc ${ $hsSwEnqReps { $sRes } } [1]) } eq # InstType pair..
           (uc ${ $hsOnlineEnqueues { $sEnqRes } } [2] ))          && # is a match..
           ( ${ $hsSwEnqReps { $sRes } } [0]                  eq # SAPSID..
             ${ $hsOnlineEnqueues { $sEnqRes } } [-2] ) ) {      # matches..
  
        $sTempNode = ${ $hsOnlineEnqueues { $sEnqRes } } [-1];   # Get the online node..
  
        @lsValidNodes = grep {                                  # Find the node other than..
                           ! /^($sTempNode|$sFaultedNode)$/o &&  # ENQUEUE online or faulted..
                           ! $hsHasysFault{$_}                   # nodes faulted..
                         } @lsValidNodes;                        # From all posibilities..
  
      &logInfo ( 20473,
      "<%s> Valid nodes for ENQREP failover targets [%s]",
      $sMyName, join ( ' ', @lsValidNodes ) );
    
        ( $sFailHost ) = @lsValidNodes if (scalar @lsValidNodes);

        if ( ( $sFailHost ne $sSysCopy ) && ( grep { /\b$sSysCopy\b/ } @lsValidNodes )) { 
          &logInfo ( 20474, "<%s> Failover host set from [%s] to [%s].", $sMyName,$sFailHost,$sSysCopy);
          $sFailHost = $sSysCopy;
        } 
        #---------------------------------------------------
        # Check if we are in a position to failover safely..
        #---------------------------------------------------
        if ( $sFaultedNode ne 'UNDEF' || $sReason ne 'MANUAL' ||
             ( $sFaultedNode eq 'UNDEF' && $sReason eq 'MANUAL' && (scalar @lsHasysFault || $sTempNode eq $sSys)) ) {
          if ( defined $sFailHost ) {
            &logInfo ( 20310,
            "<%s> Rejecting ENQUEUE node [%s]. Failing over group [%s] to node [%s]",
            $sMyName, $sTempNode, $sGrp, $sFailHost );
          }
          else { # Prevent onlining..
            &logWarn ( 20390,
            "<%s> Could not locate any node to online Group [%s]. Waiting to find a node..",
            $sMyName, $sGrp );
            next; # try again after some buffer time...
          }
        }

        #-------------------------------
        # Online group to default node..
        #-------------------------------
        if ( SAPNW::OnlineGrp ( $sGrp,
                                     $sFailHost ) ) {
          &logInfo ( 20312,
          "<%s> Issued online of Group [%s] on node [%s]",
          $sMyName, $sGrp, $sFailHost );
          return $iReturnCode; # Success..
        }
        else {
          &logError ( 20313,
          "<%s> Could not issue online of Group [%s] on node [%s]",
          $sMyName, $sGrp, $sFailHost );
          return 0; # Error..
        }
      }
      else {
        &logTrace ( 20314,
        "<%s> Failover group [%s] does not have any online ENQ(UEUE|REP) variant",
        $sMyName, $sGrp );
        next; # try again after some buffer time...
      }
    }# end of ERS processing **************
    else {
      &logInfo ( 20315,
      "<%s> Failover group [%s] does not have any ENQ(UEUE|REP) variant",
      $sMyName, $sGrp );
      return 0; # Error..
    }

    #------------------------------------------------
    # Validate if we have any online ENQREP variant..
    #------------------------------------------------
    &logTrace ( 20316,
    "<%s> +----- ENQREP online node and group details -----+", $sMyName );
  
    #----------------------------------------------
    # Remove ENQREPs whose EnqRepSrv aren't found..
    #----------------------------------------------
    foreach $sKey ( keys %hsEnqReps ) {
      if ( ! $hsEnqueues { ${$hsEnqReps { $sKey } }[1] } ) { # Not in passed group..
        &logTrace ( 20317,
        "<%s> EnqResName [%s] resource not found in Grp [%s].
        Marked resource [%s] invalid",
        $sMyName, ${$hsEnqReps { $sKey } }[1], $sGrp, $sKey );

        delete $hsEnqReps { $sKey };                        # Then delete..
      }
      else {
        ++ $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Else store online node..
        ++ $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Else store group..

        &logTrace ( 20318,
        "<%s> Online node [%s] Group [%s]",
        $sMyName, ${$hsEnqReps{$sKey}}[-1], ${$hsEnqReps{$sKey}}[0] );
      }
    }

    #------------------------------------------------
    # Validate if we have any online ENQREP variant..
    #------------------------------------------------
    if ( ! keys %hsEnqReps && $sReason ne 'MANUAL' ) {
      $sFailHost = $sSys; # Fail-over to the default node..
      &logCritical ( 20319,
      "<%s> Corresponding ENQREP server(s) not found ONLINE in cluster.
      Onlining group [%s] on node [%s]", $sMyName, $sGrp, $sFailHost );

      #-------------------------------
      # Online group to default node..
      #-------------------------------
      if ( SAPNW::OnlineGrp ( $sGrp,
                                   $sFailHost ) ) {
        &logInfo ( 20320,
        "<%s> Issued online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return $iReturnCode; # Success..
      }
      else {
        &logError ( 20321,
        "<%s> Could not issue online of Group [%s] on node [%s]",
        $sMyName, $sGrp, $sFailHost );
        return 0; # Error..
      }
    }

    #-----------------------------------------------
    # Check if we have ENQREPs in different groups..
    #-----------------------------------------------
    if ( keys %hsEnqRepGrps > 1 ) {
      &logCritical ( 20322,
      "<%s> ENQREP configured in different service groups [%s]",
      $sMyName, join ( ' ', keys %hsEnqRepGrps ) );
    }

    #----------------------------------------------
    # Now check if we have a unique failover node..
    #----------------------------------------------
    if ( keys %hsEnqRepHosts > 1 ) {
      &logCritical ( 20323,
      "<%s> ENQREP found online on different nodes [%s]",
      $sMyName, join ( ' ', keys %hsEnqRepHosts ) );

      #------------------------------------------
      # Delete entries for all other than ENQREP..
      #------------------------------------------
      foreach $sKey ( keys %hsEnqReps ) {
        if ( (uc (${$hsEnqReps { $sKey } }[2])) ne 'ENQREP' ) {
          &logError ( 20324,
          "<%s> Marking resource [%s] with InstType [%s] as invalid",
          $sMyName, $sKey, ${$hsEnqReps { $sKey } }[2] );

          #--------------------------
          # Delete in correct order..
          #--------------------------
          delete $hsEnqRepHosts { ${$hsEnqReps { $sKey } }[-1] }; # Delete node first..
          delete $hsEnqRepGrps  { ${$hsEnqReps { $sKey } }[0] };  # Delete group next..
          delete $hsEnqReps { $sKey };                            # Now the resource..
        }
      }
    }

    #-------------------------
    # Find out failover node..
    #-------------------------
    foreach $sKey ( keys %hsEnqReps ) {
      &logTrace ( 20325,
      "<%s> EnqRep server [%s] details:
      Group [%s] EnqSrvRes [%s] InstType [%s] SAPSID [%s] online on [%s]",
      $sMyName, $sKey,
      ${ $hsEnqReps { $sKey } }[0], ${ $hsEnqReps { $sKey } }[1],
      ${ $hsEnqReps { $sKey } }[2], ${ $hsEnqReps { $sKey } }[3],
      ${ $hsEnqReps { $sKey } }[4] );

      #-----------------------------------------------------------
      # Enqueue has been defined in EnqResName of ENQREP variant..
      #-----------------------------------------------------------
      if ( scalar @{ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } } ) { # EnqResName is correct..
        if ( ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0] ne    # SAPSID of ENQUEUE variant..
                              ${$hsEnqReps { $sKey }}[3] ) {          # ne SAPSID of ENQREP..
          &logError ( 20326,
          "<%s> SAPSID [%s] of ENQREP resource [%s] differs from
          SAPSID [%s] of ENQUEUE resource [%s]",
          $sMyName, ${$hsEnqReps { $sKey }}[3], $sKey,
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
          ${ $hsEnqReps { $sKey } }[1] );
        }
        elsif ( uc ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1] ne # ENQUEUE variant..
                $hsInstMatch { uc ${$hsEnqReps {$sKey }}[2] } ) {        # not correct match..
          &logError ( 20327,
          "<%s> InstType [%s] of ENQREP resource [%s] not a match for
          InstType [%s] of ENQUEUE resource [%s]", $sMyName,
          ${$hsEnqReps { $sKey }}[2], $sKey,
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
          ${ $hsEnqReps { $sKey } }[1] );
        }
        else { # All okay..
          &logTrace ( 20328,
          "<%s> EnqSrvResName [%s] matches with:
          ENQUEUE SAPSID [%s] InsType [%s] Group [%s] ProcMon [%s]", 
          $sMyName, ${$hsEnqReps { $sKey }}[1], 
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[0],
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[1],
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[2],
          ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[3] );
  
          #-----------------------------------------------------
          # Failover to ENQREP's node, if not the FAULTED node..
          #-----------------------------------------------------
          $sFailHost = ${ $hsEnqReps { $sKey } }[-1] eq $sFaultedNode ? $sSysCopy :
                       ${ $hsEnqReps { $sKey } }[-1];
          
          my $sTmpProcMon = ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[3];
          if ( $sTmpProcMon =~ /enq/ ) { # only for ENSA2 

            # hagrp -state S4H_ASCS_SG -localclus

            $sCmd = $sCmdHagrp;
            $sCmdArgs = "-state " . ${ $hsEnqueues { ${$hsEnqReps { $sKey }}[1] } }[2] . " -localclus";

            @lsHasysOut = readpipe ( "$sCmd $sCmdArgs" );
            shift @lsHasysOut;
            $iNumClusNodes = scalar @lsHasysOut; 

            if ($iNumClusNodes < 3) { # Need to check recommendation with SAP for 2 nodes
              &logInfo ( 20470, "<%s> Failover host not updated further. Service group configured on less than 3 systems.", $sMyName);
            }
            elsif ( $sFailHost eq ${ $hsEnqReps { $sKey } }[-1] ) {
              # We have chosen to failover on ERS node
              # skip the ERS node and see if there is any other valid node existing 
              # Note we should not end up going back to same node (in case of switch operation)... ***
              my $sOldFailoverHost = $sFailHost; # store old failover host
              
              @lsValidNodes = grep {                                  # Find the node other than..
                                 ! /^$sOldFailoverHost$/ &&           # ERS is ONLINE here
                                 ! $hsHasysFault{$_}                   # nodes faulted..
                               } @lsValidNodes;                        # From all posibilities..

              if ( grep { /\b$sSysCopy\b/ } @lsValidNodes ) { 
                $sFailHost = $sSysCopy;
                &logInfo ( 20464, "<%s> Failover host reset with [%s].", $sMyName,$sFailHost);
              } 
              elsif (scalar @lsValidNodes) {
                if ( ( (scalar @lsValidNodes) == 1) && ($lsValidNodes[-1] eq $sFaultedNode ) ) {
                  # This is a bug in vcs_engine, reason doesnt come correctly, so handle (faulted == validnodes && only single node)
                  # Only 1 node is valid and that node is the Source node / Faulted node, dont choose it
                  # With only 2 nodes in cluster where ENQ_SG is not faulted, this will choose ERS node
                  &logInfo ( 20465, "<%s> Failover host not updated preonline Reason [%s] ValidNodes [%s].", $sMyName,$sReason, (scalar @lsValidNodes) );
                } 
                elsif ( ( $sReason ne 'MANUAL' ) || ( (scalar @lsValidNodes) >= 1) ) {       
                  if ((scalar @lsValidNodes) > 1) {
                    my ( $sEnqueueResName ) =  keys %hsEnqueues ; # Take one..
                    $sCmdArgs = "-value " . $sEnqueueResName . " LastOnline -localclus";
                    my $sLastOnlineNode = readpipe ( "$sCmdHares $sCmdArgs" );
                    chomp $sLastOnlineNode;
                    &logInfo ( 20475, "<%s> Failover Reason [%s] ValidNodes [%s].", $sMyName, $sReason, join( ' ', @lsValidNodes ) );
                   
                    @lsValidNodes = grep {                                  # Find the node other than..
                                   ! /^$sLastOnlineNode$/                   # Resource was LastOnline here
                                 } @lsValidNodes;                           # From all posibilities..
                  }
                  $sFailHost = $lsValidNodes[-1]; 
                  &logInfo ( 20466, "<%s> Node running ENQR is not chosen. Failover host updated with [%s] ", $sMyName,$sFailHost);
                  &logInfo ( 20467, "<%s> Failover Reason [%s] ValidNodes [%s].", $sMyName, $sReason, join( ' ', @lsValidNodes ) );
                } 
                else {
                  &logInfo ( 20468, "<%s> Failover host not updated preonline Reason [%s] ValidNodes [%s].", $sMyName,$sReason, (scalar @lsValidNodes) );
                }
              } 
              else {
                &logInfo ( 20469, "<%s> Failover host not updated with any free node.", $sMyName);
              }
            }
          }
        }
      }
      else { # EnqResName is wrong..
        &logError ( 20329,
        "<%s> EnqResName [%s] value is in-correct for ENQREP resource [%s]",
        $sMyName,  ${$hsEnqReps { $sKey }}[1], $sKey );
      }
    }

    &logInfo ( 20330, "<%s> Failing over Group [%s] to node [%s]",
    $sMyName, $sGrp, $sFailHost );

    #-------------------------------
    # Online group to default node..
    #-------------------------------
    if ( SAPNW::OnlineGrp ( $sGrp,
                                 $sFailHost ) ) {
      &logInfo ( 20331, "<%s> Issued online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      last;
    }
    else {
      &logError ( 20332,
      "<%s> Could not issue online of Group [%s] on node [%s]",
      $sMyName, $sGrp, $sFailHost );
      $iReturnCode = 0; # Do nothing..
      last;
    }
  }
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool OnlineGrp ( IN String Grp,
                             IN String Sys );

 Description : Pre-online trigger for SAPNW

 Parameters :

 $sGrp                Cluster group to online
 $sSys                System where the group is asked to be made online

 Returns : boolean value in the set of {0,1} where

           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub OnlineGrp {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sGrp,
       $sSys ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20333, "Subroutine <%s> called with:
  Grp             [%s]
  Sys             [%s]",

  $sMyName, $sGrp, $sSys );

  #-------------------------------------------
  # Set return code optimistically to online..
  #-------------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sCmd = $sCmdHagrp; # Get global..
  my @lsCmdArgs = ( "-online", "-nopre", "$sGrp", "-sys", "$sSys" );
  my $iExitCode = undef;

  &logTrace ( 20334, "<%s> Executing Command [%s] with arguments [%s]",
  $sMyName, $sCmd, join( ' ', @lsCmdArgs ) );

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ! -e $sCmd ) {
    &logError ( 20335,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20336,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = 0; # Failure..
  }
  else { # Execute..
    &logTrace ( 20337,
    "<%s> Executing Cmd [%s] with args [%s]",
    $sMyName, $sCmd, join ( ' ', @lsCmdArgs ) );

    system ( $sCmd, @lsCmdArgs ); # Execute..

    #-------------------------
    # Check execution status..
    #-------------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20338,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20339,
      "<%s> Cmd [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, join ( ' ', @lsCmdArgs ), $iExitCode );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int StartSapService ( IN String InstProfile,
                               IN String SAPAdmin, 
                               IN String EnvFile,
                               IN String UseSystemD );

 Description :  Start the sapstartsrv process if found to be offline. 

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $bUseSystemD               When enabled, resource comes Online in system.slice

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub StartSapService {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sSAPAdmin,
       $sEnvFile,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20340, "Subroutine <%s> called with:
  InstProfile              [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  UseSystemD               [%s]",

  $sMyName, $sInstProfile, $sSAPAdmin, $sEnvFile, $bUseSystemD );

  my ( $sSAPSID, $sInstName, $sSAPHost ,$sSAPHome,@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";


  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );

  #------------------------------------------------
  # Define sapcontrol command and args variables..
  #------------------------------------------------
  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  $sCommand = "$sSAPHome/sapcontrol";
  $sCommandArgs = "-nr $iInstID -function StartService $sSAPSID";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20341,
  "<%s> Starting SAPNW instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  if ( $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd() ) {
    if ( ($sEnvFile =~ /\S/ )      &&         # EnvFile is specified 
         ($::gsSystemDEnvFile !~ /\S/)) {        # SystemD-compliant EnvFile is not specified 
      $::gsSystemDEnvFile = $sEnvFile;
      &logTrace (20458, "<%s> EnvFile[%s] is specified, but SystemD-compliant environment file is not specified in attribute SystemDEnvFile. 
                 The agent will use [%s] as the value for directive EnvironmentFile in the service file.",
                 $sMyName, $sEnvFile, $sEnvFile);
    }
  }
    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20035,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ($bUseSystemD) {
      &logTrace ( 20403,
      "<%s> SYSTEMD is detected", $sMyName );
      if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartService", undef ) ) {
        &logTrace ( 20404,
          "<%s> Start  command [%s] with args [%s] executed successfully",
          $sMyName, $sCommand, $sCommandArgs );
      }
      else {
        &logError ( 20407, "<%s>: Start command [%s] with args [%s] failed to execute",
          $sMyName, $sCommand, $sCommandArgs);
      }
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin,
                            $sEnvFile,
                            $sCommand,
                            $sCommandArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdout ) ) {
      &logTrace ( 20343,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  else {
    &logError ( 20406,
    "<%s> Cmd [%s] with args [%s] failed to execute.",
    $sMyName, $sCommand, $sCommandArgs );
  }
  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;
  $iTimeout = $iTillWhen - time ;
  #------------------------------------------------
  # Define sapcontrol command and args variables..
  #------------------------------------------------
  $sCommand = "$sSAPHome/sapcontrol";
  $sCommandArgs = "-nr $iInstID -function WaitforServiceStarted $iTimeout 1";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20344,
  "<%s> Waiting for Service to be started for SAPNW instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20345,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ($bUseSystemD) {
    &logTrace ( 20453,
    "<%s> SYSTEMD is detected", $sMyName );

    if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartServiceWait", undef ) ) {
      &logTrace ( 20454,
          "<%s> Start  command [%s] with args [%s] executed successfully",
          $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20455, "<%s>: Start command [%s] with args [%s] failed to execute",
          $sMyName, $sCommand, $sCommandArgs);
    }
  }
  elsif ( $iReturnCode =
    SAPNW::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
    &logTrace ( 20346,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20406,
    "<%s> Cmd [%s] with args [%s] failed to execute.",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return the error code.
  #-------------------
  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int StartERSInstance ( IN String InstProfile,
                               IN String SAPAdmin, 
                               IN String EnvFile,
                               IN Bool UseSystemD );

 Description :  Start the er process if found to be offline. 

 Parameters :

 $sInstProfile              Instance profile file for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $sEnvFile                  SAP Environment file
 $bUseSystemD               When enabled, resource comes Online in system.slice 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub StartERSInstance {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInstProfile,
       $sSAPAdmin,
       $sEnvFile,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20347, "Subroutine <%s> called with:
  InstProfile              [%s]
  SAPAdmin                 [%s]
  EnvFile                  [%s]
  UseSystemD               [%s]",

  $sMyName, $sInstProfile, $sSAPAdmin, $sEnvFile, $bUseSystemD );

  my ( $sSAPSID, $sInstName, $sSAPHost ,$sSAPHome,@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  $sSAPHome = "/usr/sap/" . $sSAPSID . "/" . $sInstName . "/exe";


  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPNW::GetEngineTimeout ( 1 );
  $iTimeout = $iTillWhen - time ;
  #------------------------------------------------
  # Define sapcontrol command and args variables..
  #------------------------------------------------
  my $iInstID = ();
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;

  $sCommand = "$sSAPHome/sapcontrol";
  $sCommandArgs = "-nr $iInstID -function StartWait $iTimeout 1";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20348,
  "<%s> Starting SAPNW instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  if ( $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd() ) {
    if ( ($sEnvFile =~ /\S/ )      &&         # EnvFile is specified 
         ($::gsSystemDEnvFile !~ /\S/)) {        # SystemD-compliant EnvFile is not specified 
      $::gsSystemDEnvFile = $sEnvFile;
      &logTrace (20459, "<%s> EnvFile[%s] is specified, but SystemD-compliant environment file is not specified in attribute SystemDEnvFile. 
                 The agent will use [%s] as the value for directive EnvironmentFile in the service file.",
                 $sMyName, $sEnvFile, $sEnvFile);
    }
  }
    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20035,
      "<%s> Insufficient time to continue boot up. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ($bUseSystemD) {
      &logTrace ( 20403,
      "<%s> SYSTEMD is detected", $sMyName );
      if ( $iReturnCode = SystemD::RunSystemdWithEnv($sSAPAdmin, $sCommand, $sCommandArgs, $::gsSystemDEnvFile, $iTimeout, "StartWait", undef ) ) {
        &logTrace ( 20404,
            "<%s> Start  command [%s] with args [%s] executed successfully",
            $sMyName, $sCommand, $sCommandArgs );
      }
      else {
         &logError ( 20407, "<%s>: Start command [%s] with args [%s] failed to execute",
            $sMyName, $sCommand, $sCommandArgs);
      } 
    }
    elsif ( $iReturnCode =
      SAPNW::ExecLocalCmd ( $sSAPAdmin,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
      &logTrace ( 20350,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  else {
    &logError ( 20406,
    "<%s> Cmd [%s] with args [%s] failed to execute.",
    $sMyName, $sCommand, $sCommandArgs );
  }
    
  #-------------------
  # Return the error code..
  #-------------------

  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int GetValueOfAttribute (IN String Attribute);

 Description : Retrieves input attribute value. 
 
 Parameters  :
          Attribute name for which value needs to be retrieved.

 Returns:  Value of input Attribute for SAPNW resource.
 -------------------------------------------------------------------------

=cut

sub GetValueOfAttribute {
  #----------------
  # Get Arguments..
  #----------------
  my ($sAttribute) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace( 20351 , "Subroutine <%s> called with [%s].",
  $sMyName, $sAttribute  );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs,
       $iValue, $iExitCode ) = ();

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #-----------------------------------------
  # Form hares command for retrieving input Attribute.
  #-----------------------------------------
  $sCommand = $sCmdHares; # Get global..
  $sCommandArgs = "-value $sResourceName $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  $iValue = 0;
  ( $iValue = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20352,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iValue ) { # Undefined value..
    &logTrace ( 20353, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20354, "<%s> Got [%s] as [%s] from engine",
    $sMyName, $sAttribute, $iValue );
  }

  #-----------------------------------------
  # Form hatype command for retrieving input Attribute.
  #-----------------------------------------
  unless ( $iValue ) {
    $sCommand = $sCmdHatype; # Get global..
    $sResourceType = "SAPNW" if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iValue = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20355,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iValue ) { # Empty value..
      &logTrace ( 20356, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20357, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iValue );
    }
  }

  return ( $iValue ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : int CheckENQLockTransfer ( IN String EnvFile,
                                     IN String ProcMon,
                                     IN String InstProfile,
                                     IN String InstType,
                                     IN String SAPAdmin );

 Description : Verifies whether the enqueue server has acquired all locks from enqueue replication server using enq_admin command

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sProcMon                  List of processes to be monitored
 $sInstProfile              Instance profile file for the SAP Instance
 $sInstType                 Instance type for the SAP Instance
 $sSAPAdmin                 SAP System Administrator
 $sEnqSrvResName            VCS resource name for Enqueue Server
 
 Returns  : Integer value as defined below:

 1  Enqueue replication server service group switch issued.
 0  No service group level operation performed.

 -------------------------------------------------------------------------

=cut

sub CheckENQLockTransfer {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sProcMon,
       $sInstProfile,
       $sInstType,
       $sSAPAdmin,
       $sEnqSrvResName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20438, "Subroutine <%s> called with:
  EnvFile                  [%s]
  ProcMon                  [%s]
  InstProfile              [%s]
  InstType                 [%s]
  SAPAdmin                 [%s]
  EnqSrvResName            [%s]",
  $sMyName, $sEnvFile, $sProcMon, $sInstProfile, $sInstType, $sSAPAdmin, $sEnqSrvResName );  

  #--------------------------------------------
  # Hopefully service group switch not required
  #--------------------------------------------
  my $iReturnCode = 0; 

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  my ($sSAPSID, $sInstName, $iInstID, $sSAPHost, $sSAPHome) = ();
  my (@lsWords) = ();

  (@lsWords) = split( /\//, $sInstProfile);
  ($sSAPSID, $sInstName, $sSAPHost) = split(/_/, $lsWords[-1]);
  ($iInstID ) = $sInstName =~ /(\d{2})$/o;
  $sSAPHome = "/usr/sap/$sSAPSID/$sInstName/exe";

  if( $sEnqSrvResName ne "" ) {
    
    # find current node name
    my $sHaresCmd = $sCmdHares;
    my $sHasyscmd = $sCmdHasys;
    my $sHasyscmdArgs = " -nodeid ";
    my $sHasysOut = readpipe ( "$sHasyscmd $sHasyscmdArgs" );
    chomp $sHasysOut;
    $sHasysOut =~ s/.*Node name: //;
    $sHasysOut =~ s/ Node ID: .*//;
    my $sNodeName = $sHasysOut;

    # check if "enq" resource is also online on the same node 
    # example: hares -state S4H_ASCS_Res -sys iar73003 
    my $sHaresCmdArgs = " -state $sEnqSrvResName  -sys $sNodeName " ;
    my $sHaCmdOut = readpipe ( "$sHaresCmd $sHaresCmdArgs" );
    chomp $sHaCmdOut;
    my $sENQUEUE_ResState = $sHaCmdOut; 
    #return 0 if ( $iExitCode = $? >> 8 );

    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };

    # find the SG name for enqr
    $sHaresCmdArgs = "-value $sResourceName Group"; # This is ERS service group
    $sHaCmdOut = readpipe ( "$sHaresCmd $sHaresCmdArgs" );
    chomp $sHaCmdOut;
    my $sENQRGroup = $sHaCmdOut; 
     

    # check whether ENQR group is already ONLINE on current node
    # If yes, then switch (don't block the first ONLINE of ENQR group)
    my $sHagrpCmd = $sCmdHagrp;
    # example: hagrp -state S4H_ENREP_SG -sys iar73003
    my $sHagrpCmdArgs = " -state $sENQRGroup  -sys $sNodeName " ;
    $sHaCmdOut = readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
    chomp $sHaCmdOut;
    my $sENQRGroupState = $sHaCmdOut; 
    if ( "ONLINE" eq $sENQRGroupState ) {

      if ( "ONLINE" eq $sENQUEUE_ResState ) {
        # We find out the group name and switch it. We can also check the group state (for ONLINE, PARTIAL_ONLINE cases)
     
        $sHagrpCmdArgs = " -state $sENQRGroup  " ;
        my @lsHasysOut = readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
        shift @lsHasysOut;

        # Sample output
        #/opt/VRTS/bin/hagrp -state  S4H_ENREP_SG -localclus
        #Group       Attribute             System     Value
        #S4H_ENREP_SG State                 iar73003   |PARTIAL|
        #S4H_ENREP_SG State                 iar73004   |OFFLINE|

        my @lsHasysRunning = grep {
                        ( s/\n//  || 1 ) &&
                        ( s/.*:// || 1 ) &&   # for DR ...
                        ( s/^$sENQRGroup.*State\s+(\S+).*OFFLINE.$/$1/ ) 
                      } @lsHasysOut;

        my $sTarget = $lsHasysRunning[0];
        my ( $sHasysCmdArgs, $sHaSysCmdOut, $sRunningSys ) = ();

        foreach $sRunningSys ( @lsHasysRunning ) { 
          $sHasysCmdArgs = " -state $sRunningSys " ;
          $sHaSysCmdOut = readpipe ( "$sCmdHasys $sHasysCmdArgs" );
          if ( $iExitCode = $? >> 8 ) {
            &logError ( 20460,
            "<%s> Cmd [%s] with args [%s] failed to execute.",
            $sMyName, $sCmdHasys, $sHasysCmdArgs );
          }
          else {
            chomp $sHaSysCmdOut ; 
            if ( $sHaSysCmdOut eq "RUNNING" ) {
              $sTarget = $sRunningSys ;
              last;   
            }
          }  
        }
        &logTrace ( 20441, "<%s> target [%s].", $sMyName, $sTarget );
        if (( defined $sTarget) && ( $sTarget =~ /\S+/ )) {


          #-----------------------
          # Set command and args..
          #-----------------------
          $sCommand = "$sSAPHome/enq_admin";
          $sCommandArgs = " --replication_state pf=$sInstProfile ";
          my $iTimeout = 30;
 
          my $bIsLockTransferComplete = 0;
          my $iEnqAdminReturnCode = SAPNW::ExecEnqAdmin ( $sEnvFile,
                                                          $sSAPAdmin,
                                                          $sCommand,
                                                          $sCommandArgs,
                                                          $iTimeout );
          if ( $iEnqAdminReturnCode == 0 ) {
            &logTrace ( 20444, "<%s> Enqueue replication server service group switch will be issued.", $sMyName );
          } else {
            &logTrace ( 20443, "<%s> Enqueue replication server service group switch operation not initiated.", $sMyName );
            return 0;
          }
           # found the SG name and now switch it within -localclus (ENQ / ERS should be in same cluster)
          # hagrp -switch <group> -any [-clus <cluster> | -localclus
          $sHagrpCmdArgs = " -switch $sENQRGroup  -to $sTarget " ;
          $sHaCmdOut = readpipe ( "$sHagrpCmd $sHagrpCmdArgs" );
          if ( $iExitCode = $? >> 8 ) {
            &logTrace ( 20438, "<%s> Enqueue replication server service group switch operation failed.", $sMyName );
          } else {
            &logTrace ( 20437, "<%s> Enqueue replication server service group switch issued.", $sMyName );
            $iReturnCode = 1;
            return $iReturnCode;
          }
        } else {
          &logTrace ( 20442, "<%s> Could not find suitable target for Enqueue Replication server.", $sMyName );
        }
      } else {
        &logTrace ( 20440, "<%s> Enqueue resource is not online on current node.", $sMyName );
      }
    } else {
      &logTrace ( 20439, "<%s> Enqueue replication server service group is not online on current node.", $sMyName );
    }
  }
    
  return $iReturnCode; 
} 

sub ExecEnqAdmin {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sSAPAdmin,
       $sCmd,
       $sCmdArgs,
       $iTimeout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20257, "Subroutine <%s> called with:
  SAPAdmin      [%s]
  EnvFile       [%s]
  Cmd           [%s]
  CmdArgs       [%s]
  Timeout       [%s]",

  $sMyName,
  $sSAPAdmin, $sEnvFile, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------------------------------------------------------
  # Only when lock recovery is complete, return success (0), default (-1) 
  #----------------------------------------------------------------------
  my $iReturnCode = -1;
 
  # check if locks are acquired by enq and we can move out enqr
  #
  # iar73003:s4hadm 51> enq_admin --replication_state pf=/usr/sap/S4H/SYS/profile/S4H_ERS11_s4hers
  # Enqueue Server 2
  #
  # Replication-State = ON
  #
  # 2019-04-04 14:34:02; OK; 'Replication State'; Response=304 usec
  # ===================================================================================================
  # iar73003:s4hadm 52> echo $?
  # 0
  #
  #   --replication_state
  # Gets the replication state from the server.
  # 0 = Replication is not configured or
  #     Replication is configured and is in-sync  ******** This indicates lock transfer complete & ready to switch node
  # 3 = Enqueue Server 2 can not be reached
  # 4 = Replication is configured but is NOT in-sync


  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iExitCode, @lsCmdOut ) = ();

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -e $sCmd ) {
    &logError ( 20258,
    "<%s> SAP Cmd [%s] does not exist", $sMyName, $sCmd );
    $iReturnCode = -1;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20259,
    "<%s> SAP Cmd [%s] is not executable", $sMyName, $sCmd );
    $iReturnCode = -1;
  }
  elsif ( $iTimeout < 10 ) {
    &logError ( 20260,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with args [%s]",
    $sMyName, $iTimeout, $sCmd, $sCmdArgs );
    $iReturnCode = -1;
  }
  elsif ( SAPNW::ExecLocalCmd ( $sSAPAdmin, # User context..
                                     $sEnvFile,  # Use envfile..
                                     $sCmd,
                                     $sCmdArgs,
                                     \$iExitCode,
                                     $iTimeout,
                                     \@lsCmdOut ) ) {

    #-------------------------------
    # Command execution successful..
    #-------------------------------
    &logTrace ( 20261,
    "<%s> Cmd [%s] with args [%s] ran successfully with exit code [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode );
  }
  if ( ! defined $iExitCode ) { # Time-out..
    $lsCmdOut [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = -1; # Failure..

    &logError ( 20262,
    "<%s> Cmd [%s] with args [%s] timed out in [%s] sec. Returning [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout, $iReturnCode );
  }
  elsif ( 3 == $iExitCode ) {
      &logTrace ( 20264, "<%s> Cmd [%s] with args [%s] exited with [%s].  Enqueue Server 2 can not be reached", $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      $iReturnCode = $iExitCode; # Failure..
  }
  elsif ( 4 == $iExitCode ) {  
      &logTrace ( 20264, "<%s> Cmd [%s] with args [%s] exited with [%s]. Replication is configured but is NOT in-sync", $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      $iReturnCode = $iExitCode; # Failure..
  }
  elsif ( 0 == $iExitCode ) {
      &logTrace ( 20264, "<%s> Cmd [%s] with args [%s] exited with [%s].", $sMyName, $sCmd, $sCmdArgs, $iExitCode );
      $iReturnCode = $iExitCode; # Failure..
  }
  else { # Error..
    $iReturnCode = -1;
    &logError ( 20267,
    "<%s> Cmd [%s] failed with exit code [%s]. Returning [%s]",
    $sMyName, $sCmd, $iExitCode, $iReturnCode );
  }

  return $iReturnCode;
}
1; # So PERL knows its included..
